#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2020 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
Generate the pyxtuml bridgepoint schema module from a bridgepoint metamodel.
'''

import optparse
import re
import string
import sys
import logging
import re
import xtuml

from xtuml import navigate_many as many
from xtuml import navigate_one as one
from xtuml import navigate_subtype as subtype
from xtuml import relate
from xtuml import unrelate
from xtuml import where_eq as where

from bridgepoint import ooaofooa


tmpl = string.Template("""
# encoding: utf-8
# Copyright (C) 2020 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
This file was generated by bridgepoint.gen_pyxtuml_schema.
'''

__version__ = '${version}'


classes = '''
${classes}
'''

associations = '''
${associations}
'''

indices = '''
${indices}
'''

globals = '''
${globals}
'''
""")


logger = logging.getLogger('schema_gen')


def parse_keyword(expr, keyword):
    regexp = re.compile(keyword + ":([^\n]*)")
    result = regexp.search(expr)
        
    if result:
        return result.groups()[0].strip().lower()
    else:
        return ''

    
def description_filter(inst):
    description = inst.Descrip.lower()
    keymap = dict(persistent='false',
                  translate_for_external_use='false')
                  
    for key, value in keymap.items():
        if parse_keyword(description, key) == value:
            return True
    

def mk_ooaofooa_component(filename):
    loader = ooaofooa.Loader()
    loader.filename_input(filename)
    
    m = loader.build_metamodel()

    for r_rel in m.select_many('R_REL', description_filter):
        logger.info('Filtering R%d' % r_rel.Numb)
        xtuml.delete(r_rel)

    for o_obj in m.select_many('O_OBJ', description_filter):
        logger.info('Filtering %s' % o_obj.Key_Lett)
        for r_rel in many(o_obj).R_OIR[201].R_REL[201]():
            logger.info('Filtering R%d' % r_rel.Numb)
            for r_oir in many(r_rel).R_OIR[201]():
                xtuml.unrelate(r_rel, r_oir, 201)
                
            xtuml.delete(r_rel)
            
        xtuml.delete(o_obj)

    for o_attr in m.select_many('O_ATTR', where(Name='SMspd_ID')):
        if not one(o_attr).O_RATTR[106]():
            continue
            
        for o_oida in many(o_attr).O_OIDA[105]():
            for o_rtida in many(o_oida).O_RTIDA[110]():
                xtuml.delete(o_rtida)
            xtuml.delete(o_oida)

        for o_ref in many(o_attr).O_RATTR[106].O_REF[108]():
            xtuml.delete(o_ref)
            
    c = ooaofooa.mk_component(m, None, derived_attributes=True)
    metaclass = c.find_metaclass('ACT_ACT')
    metaclass.insert_attribute(index=5, name='return_value', type_name='INTEGER')

    for o_obj in m.select_many('O_OBJ'):
        for o_attr in many(o_obj).O_ATTR[102](description_filter):
            logger.info('Filtering %s.%s' % (o_obj.Key_Lett, o_attr.Name))
            metaclass = c.find_metaclass(o_obj.Key_Lett)
            metaclass.delete_attribute(o_attr.Name)
            
    return c


def main():
    '''
    Parse argv for options and arguments, and start schema generation.
    '''
    parser = optparse.OptionParser(usage="%prog [options] <bridgepoint source path>",
                                   version=xtuml.version.complete_string,
                                   formatter=optparse.TitledHelpFormatter())
                                   
    parser.set_description(__doc__.strip())
    
    parser.add_option("-o", "--output", dest='output', metavar="PATH",
                      help="save sql schema to PATH (required)",
                      action="store", default=None)

    parser.add_option("-v", "--verbosity", dest='verbosity', action="count", 
                      help="increase debug logging level", default=2)

    
    (opts, args) = parser.parse_args()
    if len(args) == 0 or opts.output is None:
        parser.print_help()
        sys.exit(1)

    levels = {
              0: logging.ERROR,
              1: logging.WARNING,
              2: logging.INFO,
              3: logging.DEBUG,
    }
    logging.basicConfig(level=levels.get(opts.verbosity, logging.DEBUG))

    ooaofooa_path = args[0] + '/src/org.xtuml.bp.core/models/org.xtuml.bp.core/ooaofooa'
    globals_path = args[0] + '/src/org.xtuml.bp.pkg/globals/Globals.xtuml'
    version_path = args[0] + '/src/org.xtuml.bp.pkg/about.mappings'
    
    with open(globals_path, 'r') as f:
        globals_instances = f.read()
        
    m = mk_ooaofooa_component(ooaofooa_path)

    with open(version_path, 'r') as f:
        version = re.findall(r'0=([\d\.]*)\.qualifier', f.read())[0]
    
    s = tmpl.substitute(version=version,
                        classes=xtuml.serialize_classes(m).strip(),
                        associations=xtuml.serialize_associations(m).strip(),
                        indices=xtuml.serialize_unique_identifiers(m).strip(),
                        globals=globals_instances.strip())

    with open(opts.output, 'w') as f:
        f.write(s.strip())
        f.write('\n')
    
if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    main()
    



# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
Check an xtuml model for association constraint violations in its metamodel.
'''

import logging
import optparse
import sys
import xtuml


logger = logging.getLogger('consistency_check')


def pretty_to_link(inst, link):
    '''
    Create a human-readable representation of a link on the 'TO'-side
    '''
    values = ''
    prefix = ''
    metaclass = xtuml.get_metaclass(inst)

    for name, ty in metaclass.attributes:
        if name in link.key_map:
            value = getattr(inst, name)
            value = xtuml.serialize_value(value, ty)
            name = link.key_map[name]
            values += '%s%s=%s' % (prefix, name, value)
            prefix = ', '
                
    return '%s(%s)' % (link.kind, values)
        

def pretty_from_link(inst, link):
    '''
    Create a human-readable representation of a link on the 'FROM'-side
    '''
    values = ''
    prefix = ''
    metaclass = xtuml.get_metaclass(inst)
    
    for name, ty in metaclass.attributes:
        if name in link.key_map:
            value = getattr(inst, name)
            value = xtuml.serialize_value(value, ty)
            values += '%s%s=%s' % (prefix, name, value)
            prefix = ', '
                
    return '%s(%s)' % (metaclass.kind, values)


def pretty_unique_identifier(inst, identifier):
    '''
    Create a human-readable representation a unique identifier.
    '''
    values = ''
    prefix = ''
    metaclass = xtuml.get_metaclass(inst)
    
    for name, ty in metaclass.attributes:
        if name in metaclass.identifying_attributes:
            value = getattr(inst, name)
            value = xtuml.serialize_value(value, ty)
            values += '%s%s=%s' % (prefix, name, value)
            prefix = ', '
                    
    return '%s(%s)' % (identifier, values)


def check_uniqueness_constraint(m, kind=None):
    '''
    Check the model for uniqueness constraint violations.
    '''
    if kind is None:
        metaclasses = m.metaclasses.values()
    else:
        metaclasses = [m.find_metaclass(kind)]
    
    res = 0
    for metaclass in metaclasses:
        id_map = dict()
        for identifier in metaclass.indices:
            id_map[identifier] = dict()
                
        for inst in metaclass.select_many():
            # Check for null-values
            for name, ty in metaclass.attributes:
                if name not in metaclass.identifying_attributes:
                    continue
                
                value = getattr(inst, name)
                isnull = value is None
                isnull |= (ty == 'UNIQUE_ID' and not value)
                if isnull:
                    res += 1 
                    logger.warning('%s.%s is part of an identifier and is null' 
                                   % (metaclass.kind, name))

            # Check uniqueness
            for identifier in metaclass.indices:
                kwargs = dict()
                for name in metaclass.indices[identifier]:
                    kwargs[name] = getattr(inst, name)

                index_key = frozenset(kwargs.items())
                if index_key in id_map[identifier]:
                    res += 1
                    id_string = pretty_unique_identifier(inst, identifier)
                    logger.warning('uniqueness constraint violation in %s, %s' 
                                   % (metaclass.kind, id_string))

                id_map[identifier][index_key] = inst

    return res


def check_link_integrity(m, link):
    '''
    Check the model for integrity violations on an association in a particular direction.
    '''
    res = 0
    for inst in link.from_metaclass.select_many():
        q_set = list(link.navigate(inst))

        if(len(q_set) < 1 and not link.conditional) or (
          (len(q_set) > 1 and not link.many)):
            res += 1
            logger.warning('integrity violation in '
                           '%s --(%s)--> %s' % (pretty_from_link(inst, link),
                                                link.rel_id,
                                                pretty_to_link(inst, link)))
    
    return res


def check_subtype_integrity(m, super_kind, rel_id):
    '''
    Check the model for integrity violations across a subtype association.
    '''
    if isinstance(rel_id, int):
        rel_id = 'R%d' % rel_id

    res = 0
    for inst in m.select_many(super_kind):
        if not xtuml.navigate_subtype(inst, rel_id):
            res += 1
            logger.warning('integrity violation across '
                           '%s[%s]' % (super_kind, rel_id))
        
    return res


def check_association_integrity(m, rel_id=None):
    '''
    Check the model for integrity violations on association(s).
    '''
    if isinstance(rel_id, int):
        rel_id = 'R%d' % rel_id
            
    res = 0
    for ass in m.associations:
        if rel_id in [ass.rel_id, None]:
            res += check_link_integrity(m, ass.source_link)
            res += check_link_integrity(m, ass.target_link)

    return res


def main(args):
    parser = optparse.OptionParser(usage="%prog [options] <sql_file> [another_sql_file...].",
                                   version=xtuml.version.complete_string,
                                   formatter=optparse.TitledHelpFormatter())
    
    parser.set_description(__doc__.strip())
    
    parser.add_option("-r", "-R", dest="rel_ids", type='int', metavar="<number>",
                      help="limit consistency check to one or more associations",
                      action="append", default=[])

    parser.add_option("-k", dest="kinds", type='string', metavar="<key letter>",
                      help="limit check for uniqueness constraint violations to one or more classes",
                      action="append", default=[])
    
    parser.add_option("-v", "--verbosity", dest='verbosity', action="count",
                      help="increase debug logging level", default=1)
    
    (opts, args) = parser.parse_args(args)
    if len(args) == 0:
        parser.print_help()
        sys.exit(1)
        
    levels = {
              0: logging.ERROR,
              1: logging.WARNING,
              2: logging.INFO,
              3: logging.DEBUG,
    }
    logging.basicConfig(level=levels.get(opts.verbosity, logging.DEBUG))
    
    loader = xtuml.ModelLoader()
    for filename in args:
        loader.filename_input(filename)

    m = loader.build_metamodel()
    
    error = 0
    for rel_id in opts.rel_ids:
        error += xtuml.check_association_integrity(m, rel_id)
    
    if not opts.rel_ids:
        error += xtuml.check_association_integrity(m)

    for kind in opts.kinds:
        error += xtuml.check_uniqueness_constraint(m, kind)
    
    if not opts.kinds:
        error += xtuml.check_uniqueness_constraint(m)
        
    return error
    
    
if __name__ == '__main__':
    num_errors = main(sys.argv[1:])
    sys.exit(num_errors > 0)




# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
Serialize xtuml models and its schema to an sql-based file format and persist
to disk.
'''

import uuid
import logging

import xtuml


logger = logging.getLogger(__name__)


def serialize_value(value, ty):
    '''
    Serialize a value from an xtuml metamodel instance.
    '''
    ty = ty.upper()
    
    null_value = {
        'BOOLEAN'   : False,
        'INTEGER'   : 0,
        'REAL'      : 0.0,
        'STRING'    : '',
        'UNIQUE_ID' : 0
    }
    
    transfer_fn = {
        'BOOLEAN'     : lambda v: '%d' % int(v),
        'INTEGER'     : lambda v: '%d' % v,
        'REAL'        : lambda v: '%f' % v,
        'STRING'      : lambda v: "'%s'" % v.replace("'", "''"),
        'UNIQUE_ID'   : lambda v: '"%s"' % uuid.UUID(int=v)
    }

    if value is None:
        value = null_value[ty]
    
    return transfer_fn[ty](value)
    
    
def serialize_instance(instance):
    '''
    Serialize an *instance* from a metamodel.
    '''
    attr_count = 0
    metaclass = xtuml.get_metaclass(instance)
    s = 'INSERT INTO %s VALUES (' % metaclass.kind
    for name, ty in metaclass.attributes:
        value = getattr(instance, name)
            
        s += '\n    '
        s += serialize_value(value, ty)

        attr_count += 1
        if attr_count < len(metaclass.attributes):
            s += ', -- %s : %s' % (name, ty)
        else:
            s += ' -- %s : %s' % (name, ty)

    s += '\n);\n'

    return s


def serialize_instances(metamodel):
    '''
    Serialize all instances in a *metamodel*.
    '''
    s = ''
    for inst in metamodel.instances:
        s += serialize_instance(inst)
    
    return s


def serialize_association(ass):
    '''
    Serialize an xtuml metamodel association.
    '''
    s1 = '%s %s (%s)' % (ass.source_link.cardinality,
                         ass.source_link.to_metaclass.kind,
                         ', '.join(ass.source_keys))

    if ass.target_link.phrase:
        s1 += " PHRASE '%s'" % ass.target_link.phrase

    s2 = '%s %s (%s)' % (ass.target_link.cardinality,
                         ass.target_link.to_metaclass.kind,
                         ', '.join(ass.target_keys))
    
    if ass.source_link.phrase:
        s2 += " PHRASE '%s'" % ass.source_link.phrase

    return 'CREATE ROP REF_ID %s FROM %s TO %s;\n' % (ass.rel_id,
                                                      s1,
                                                      s2)


def serialize_class(Cls):
    '''
    Serialize an xtUML metamodel class.
    '''
    metaclass = xtuml.get_metaclass(Cls)
    attributes = ['%s %s' % (name, ty.upper())
                  for name, ty in metaclass.attributes]
    
    s = 'CREATE TABLE %s (\n    ' % metaclass.kind
    s += ',\n    '.join(attributes)
    s += '\n);\n'

    return s

def serialize_unique_identifiers(metamodel):
    s = ''
    for kind in sorted(metamodel.metaclasses.keys()):
        metaclass = metamodel.metaclasses[kind]
        for index_name, attribute_names in metaclass.indices.items():
            attribute_names = ', '.join(attribute_names)
            s += 'CREATE UNIQUE INDEX %s ON %s (%s);\n' % (index_name,
                                                          metaclass.kind,
                                                          attribute_names)
    return s


def serialize_classes(metamodel):
    '''
    Serialize all class definitions in a *metamodel*.
    '''
    return ''.join([serialize_class(metamodel.metaclasses[kind].clazz)
                    for kind in sorted(metamodel.metaclasses.keys())])


def serialize_associations(metamodel):
    '''
    Serialize all association definitions in a *metamodel*.
    '''
    orderby = lambda x: (x.rel_id, x.target_link.from_metaclass.kind)
    return ''.join([serialize_association(ass)
                    for ass in sorted(metamodel.associations, key=orderby)])


def serialize_schema(metamodel):
    '''
    Serialize all class and association definitions in a *metamodel*.
    '''
    return serialize_classes(metamodel) + serialize_associations(metamodel)


def serialize_database(metamodel):
    '''
    Serialize all instances, class definitions, association definitions, and
    unique identifiers  in a *metamodel*.
    '''
    schema = serialize_schema(metamodel)
    instances = serialize_instances(metamodel)
    identifiers = serialize_unique_identifiers(metamodel)
    
    return ''.join([schema, instances, identifiers])


def serialize(resource):
    '''
    Serialize some xtuml *resource*, e.g. an instance or a complete metamodel.
    '''
    if isinstance(resource, xtuml.MetaModel):
        return serialize_database(resource)

    elif isinstance(resource, type) and issubclass(resource, xtuml.Class):
        return serialize_class(resource)
    
    elif isinstance(resource, xtuml.Association):
        return serialize_association(resource)

    elif isinstance(resource, xtuml.Class):
        return serialize_instance(resource)


def persist_instances(metamodel, path, mode='w'):
    '''
    Persist all instances in a *metamodel* by serializing them and saving to a 
    *path* on disk.
    '''
    with open(path, mode) as f:
        for inst in metamodel.instances:
            s = serialize_instance(inst)
            f.write(s)


def persist_schema(metamodel, path, mode='w'):
    '''
    Persist all class and association definitions in a *metamodel* by 
    serializing them and saving to a *path* on disk.
    '''
    with open(path, mode) as f:
        for kind in sorted(metamodel.metaclasses.keys()):
            s = serialize_class(metamodel.metaclasses[kind].clazz)
            f.write(s)
            
        for ass in sorted(metamodel.associations, key=lambda x: x.rel_id):
            s = serialize_association(ass)
            f.write(s)


def persist_unique_identifiers(metamodel, path, mode='w'):
    '''
    Persist all unique identifiers in a *metamodel* by serializing them and
    saving to a *path* on disk.
    '''
    with open(path, mode) as f:
        for metaclass in metamodel.metaclasses.values():
            for index_name, attribute_names in metaclass.indices.items():
                attribute_names = ', '.join(attribute_names)
                s = 'CREATE UNIQUE INDEX %s ON %s (%s);\n' % (index_name,
                                                              metaclass.kind,
                                                              attribute_names)
                f.write(s)


def persist_database(metamodel, path, mode='w'):
    '''
    Persist all instances, class definitions and association definitions in a
    *metamodel* by serializing them and saving to a *path* on disk.
    '''
    with open(path, mode) as f:
        for kind in sorted(metamodel.metaclasses.keys()):
            metaclass = metamodel.metaclasses[kind]
            s = serialize_class(metaclass.clazz)
            f.write(s)
            
            for index_name, attribute_names in metaclass.indices.items():
                attribute_names = ', '.join(attribute_names)
                s = 'CREATE UNIQUE INDEX %s ON %s (%s);\n' % (index_name,
                                                              metaclass.kind,
                                                              attribute_names)
                f.write(s)
                
        for ass in sorted(metamodel.associations, key=lambda x: x.rel_id):
            s = serialize_association(ass)
            f.write(s)

        for inst in metamodel.instances:
            s = serialize_instance(inst)
            f.write(s)




# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
Check a bridgepoint model for association constraint violations in its 
metamodel (ooaofooa).
'''

import logging
import optparse
import sys
import xtuml
from bridgepoint import ooaofooa


logger = logging.getLogger('consistency_check')


def main(args):
    parser = optparse.OptionParser(usage="%prog [options] <model_path> [another_model_path...]",
                                   version=xtuml.version.complete_string,
                                   formatter=optparse.TitledHelpFormatter())
    
    parser.set_description(__doc__.strip())
    
    parser.add_option("-r", "-R", dest="rel_ids", type='int', metavar="<number>",
                      help="limit consistency check to one or more associations",
                      action="append", default=[])

    parser.add_option("-k", dest="kinds", type='string', metavar="<key letter>",
                      help="limit check for uniqueness constraint violations to one or more classes",
                      action="append", default=[])
    
    parser.add_option("-g", "--globals", dest="globals",
                      help="add builtin global data types automatically, e.g. boolean, integer and real",
                      action="store_true", default=False)
                      
    parser.add_option("-v", "--verbosity", dest='verbosity', action="count",
                      help="increase debug logging level", default=1)
    
    (opts, args) = parser.parse_args(args)
    if len(args) == 0:
        parser.print_help()
        sys.exit(1)
        
    levels = {
              0: logging.ERROR,
              1: logging.WARNING,
              2: logging.INFO,
              3: logging.DEBUG,
    }
    logging.basicConfig(level=levels.get(opts.verbosity, logging.DEBUG))
    
    loader = ooaofooa.Loader(load_globals=opts.globals)
    for filename in args:
        loader.filename_input(filename)

    m = loader.build_metamodel()
    
    error = 0
    for rel_id in opts.rel_ids:
        error += xtuml.check_association_integrity(m, rel_id)
    
    if not opts.rel_ids:
        error += xtuml.check_association_integrity(m)

    for kind in opts.kinds:
        error += xtuml.check_uniqueness_constraint(m, kind)
    
    if not opts.kinds:
        error += xtuml.check_uniqueness_constraint(m)
    
    return error
    

if __name__ == '__main__':
    num_errors = main(sys.argv[1:])
    sys.exit(num_errors > 0)



# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.

'''
Parser for the Object Action Language (OAL). 
Heavily inspired by: 
   - https://github.com/xtuml/mc/blob/master/mcmc/arlan/arlan.l
   - https://github.com/xtuml/mc/blob/master/mcmc/arlan/arlan.y
   - https://github.com/xtuml/bridgepoint/blob/master/src/org.xtuml.bp.als.oal/bnf/oal.bnf
'''

import os
import logging

from ply import lex
from ply import yacc
from functools import wraps


logger = logging.getLogger(__name__)


class Position(object):
    start_line = -1
    start_column = -1
    end_line = -1
    end_column = -1
    
    def __init__(self, start_line=0, start_column=0, end_line=0, end_column=0):
        self.start_line = start_line
        self.start_column = start_column
        self.end_line = end_line
        self.end_column = end_column
        
    def __str__(self):
        return '%s:%s - %s:%s' % (self.start_line, self.start_column,
                                  self.end_line, self.end_column)
    
    
class Node(object):
    position = None
    character_stream = None
    
    @property
    def children(self):
        return tuple()
    
    def __str__(self):
        return '[%s]:%s' % (self.position, self.__class__.__name__)
    
    
class BodyNode(Node):
    block = None
    
    def __init__(self, block):
        self.block = block
        
    @property
    def children(self):
        return (self.block,)
    
    
class BlockNode(Node):
    statement_list = None
    
    def __init__(self, statement_list):
        self.statement_list = statement_list
        
    @property
    def children(self):
        return (self.statement_list,)


class StatementListNode(Node):
    children = None
    
    def __init__(self):
        self.children = list()
            
    
class BreakNode(Node):
    pass


class ContinueNode(Node):
    pass


class ControlNode(Node):
    pass


class ReturnNode(Node):
    expression = None
    
    def __init__(self, expression):
        self.expression = expression
        
    @property
    def children(self):
        return (self.expression,)


class AssignmentNode(Node):
    variable_access = None
    expression = None
    
    def __init__(self, variable_access, expression):
        self.variable_access = variable_access
        self.expression = expression
        
    @property
    def children(self):
        return (self.variable_access, self.expression)
    
    
class InvocationStatementNode(Node):
    invocation = None
    
    def __init__(self, invocation):
        self.invocation = invocation

    @property
    def children(self):
        return (self.invocation,)
    

class MessageStatementNode(Node):
    invocation = None
    
    def __init__(self, invocation):
        self.invocation = invocation

    @property
    def children(self):
        return (self.invocation,)


class GenerateClassEventNode(Node):
    event_specification = None
    key_letter = None
    
    def __init__(self, event_specification, key_letter):
        self.event_specification = event_specification
        self.key_letter = key_letter
        
    @property
    def children(self):
        return (self.event_specification,)


class GenerateCreatorEventNode(Node):
    event_specification = None
    key_letter = None
    
    def __init__(self, event_specification, key_letter):
        self.event_specification = event_specification
        self.key_letter = key_letter
        
    @property
    def children(self):
        return (self.event_specification,)


class GenerateInstanceEventNode(Node):
    event_specification = None
    variable_access = None
    
    def __init__(self, event_specification, variable_access):
        self.event_specification = event_specification
        self.variable_access = variable_access
        
    @property
    def children(self):
        return (self.event_specification, self.variable_access)


class CreateClassEventNode(Node):
    variable_name = None
    event_specification = None
    key_letter = None
    
    def __init__(self, variable_name, event_specification, key_letter):
        self.variable_name = variable_name
        self.event_specification = event_specification
        self.key_letter = key_letter
        
    @property
    def children(self):
        return (self.event_specification,)


class CreateCreatorEventNode(Node):
    variable_name = None
    event_specification = None
    key_letter = None
    
    def __init__(self, variable_name, event_specification, key_letter):
        self.variable_name = variable_name
        self.event_specification = event_specification
        self.key_letter = key_letter
        
    @property
    def children(self):
        return (self.event_specification,)


class CreateInstanceEventNode(Node):
    variable_name = None
    event_specification = None
    to_variable_access = None
    
    def __init__(self, variable_name, event_specification, to_variable_access):
        self.variable_name = variable_name
        self.event_specification = event_specification
        self.to_variable_access = to_variable_access
        
    @property
    def children(self):
        return (self.event_specification, self.to_variable_access)


class GeneratePreexistingNode(Node):
    variable_access = None
    
    def __init__(self, variable_access):
        self.variable_access = variable_access
        
    @property
    def children(self):
        return (self.variable_access,)
    
    
    
class CreateObjectNode(Node):
    variable_name = None
    key_letter = None
    
    def __init__(self, variable_name, key_letter):
        self.variable_name = variable_name
        self.key_letter = key_letter


class CreateObjectNoVariableNode(Node):
    key_letter = None
    
    def __init__(self, key_letter):
        self.key_letter = key_letter
        

class DeleteNode(Node):
    variable_name = None
    
    def __init__(self, variable_name):
        self.variable_name = variable_name

    @property
    def children(self):
        return (self.variable_name,)


class CreateEventNode(Node):
    variable_name = None
    event_spec = None
    
    def __init__(self, variable_name, event_spec):
        self.variable_name = variable_name
        self.event_spec = event_spec
        
    @property
    def children(self):
        return (self.event_spec,)


class EventSpecNode(Node):
    identifier = None
    meaning = None
    event_data = None
    
    def __init__(self, identifier, meaning, event_data):
        self.identifier = identifier
        self.meaning = meaning
        self.event_data = event_data
    
    @property
    def children(self):
        return (self.event_data,)


class EventDataListNode(Node):
    children = None
    
    def __init__(self):
        self.children = list()
    

class EventDataItemNode(Node):
    name = None
    expression = None
    
    def __init__(self, name, expression):
        self.name = name
        self.expression = expression
        
    @property
    def children(self):
        return (self.expression,)


class ForEachNode(Node):
    instance_variable_name = None
    set_variable_name = None
    block = None
    
    def __init__(self, instance_variable_name, set_variable_name, block):
        self.instance_variable_name = instance_variable_name
        self.set_variable_name = set_variable_name
        self.block = block
        
    @property
    def children(self):
        return (self.block,)
    
    
class WhileNode(Node):
    expression = None
    block = None
    
    def __init__(self, expression, block):
        self.expression = expression
        self.block = block
        
    @property
    def children(self):
        return (self.expression, self.block)


class IfNode(Node):
    expression = None
    block = None
    else_clause = None
    elif_list = None
    
    def __init__(self, expression, block, elif_list, else_clause):
        self.expression = expression
        self.block = block
        self.elif_list = elif_list
        self.else_clause = else_clause
        
    @property
    def children(self):
        return (self.expression, self.block, self.elif_list, self.else_clause)
    
    
class ElIfListNode(Node):
    children = None
    
    def __init__(self):
        self.children = list()
        
        
class ElIfNode(Node):
    expression = None
    block = None
    
    def __init__(self, expression, block):
        self.expression = expression
        self.block = block

    @property
    def children(self):
        return (self.expression, self.block)
    
    
class ElseNode(Node):
    block = None
    
    def __init__(self, block):
        self.block = block
        
    @property
    def children(self):
        return (self.block,)


class RelateNode(Node):
    from_variable_name = None
    to_variable_name = None
    rel_id = None
    phrase = None
    
    def __init__(self, from_variable_name, to_variable_name, rel_id, phrase):
        self.from_variable_name = from_variable_name
        self.to_variable_name = to_variable_name
        self.rel_id = rel_id
        self.phrase = phrase or ''
    

class RelateUsingNode(Node):
    from_variable_name = None
    to_variable_name = None
    using_variable_name = None
    rel_id = None
    phrase = None
    
    def __init__(self, from_variable_name, to_variable_name, rel_id, phrase, using_variable_name):
        self.from_variable_name = from_variable_name
        self.to_variable_name = to_variable_name
        self.using_variable_name = using_variable_name
        self.rel_id = rel_id
        self.phrase = phrase or ''


class UnrelateNode(Node):
    from_variable_name = None
    to_variable_name = None
    rel_id = None
    phrase = None
    
    def __init__(self, from_variable_name, to_variable_name, rel_id, phrase):
        self.from_variable_name = from_variable_name
        self.to_variable_name = to_variable_name
        self.rel_id = rel_id
        self.phrase = phrase or ''


class UnrelateUsingNode(Node):
    from_variable_name = None
    to_variable_name = None
    using_variable_name = None
    rel_id = None
    phrase = None
    
    def __init__(self, from_variable_name, to_variable_name, rel_id, phrase, using_variable_name):
        self.from_variable_name = from_variable_name
        self.to_variable_name = to_variable_name
        self.using_variable_name = using_variable_name
        self.rel_id = rel_id
        self.phrase = phrase or ''
    

class SelectRelatedNode(Node):
    cardinality = None
    variable_name = None
    handle = None
    navigation_chain = None

    
    def __init__(self, cardinality, variable_name, handle, navigation_chain):
        self.cardinality = cardinality
        self.variable_name = variable_name
        self.handle = handle
        self.navigation_chain = navigation_chain
    
    @property
    def many(self):
        return self.cardinality.lower() == 'many'
    
    @property
    def children(self):
        return (self.handle, self.navigation_chain)


class SelectRelatedWhereNode(Node):
    cardinality = None
    variable_name = None
    handle = None
    navigation_chain = None
    where_clause = None
    
    def __init__(self, cardinality, variable_name, handle, navigation_chain, where_clause):
        self.cardinality = cardinality
        self.variable_name = variable_name
        self.handle = handle
        self.navigation_chain = navigation_chain
        self.where_clause = where_clause
    
    @property
    def many(self):
        return self.cardinality.lower() == 'many'
    
    @property
    def children(self):
        return (self.handle, self.navigation_chain, self.where_clause)


class NavigationListNode(Node):
    children = None
    
    def __init__(self):
        self.children = list()
        

class NavigationStepNode(Node):
    key_letter = None
    rel_id = None
    phrase = None
    
    def __init__(self, key_letter, rel_id, phrase):
        self.key_letter = key_letter
        self.rel_id = rel_id
        self.phrase = phrase or ''
    

class SelectFromNode(Node):
    cardinality = None
    variable_name = None
    key_letter = None
    
    def __init__(self, cardinality, variable_name, key_letter):
        self.cardinality = cardinality
        self.variable_name = variable_name
        self.key_letter = key_letter
    
    @property
    def many(self):
        return self.cardinality.lower() == 'many'
    
    
class SelectFromWhereNode(Node):
    cardinality = None
    variable_name = None
    key_letter = None
    where_clause = None
    
    def __init__(self, cardinality, variable_name, key_letter, where_clause):
        self.cardinality = cardinality
        self.variable_name = variable_name
        self.key_letter = key_letter
        self.where_clause = where_clause
        
    @property
    def many(self):
        return self.cardinality.lower() == 'many'
    
    @property
    def children(self):
        return (self.where_clause,)


class InstanceInvocationNode(Node):
    handle = None
    action_name = None
    parameter_list = None
    
    def __init__(self, handle, action_name, parameter_list):
        self.handle = handle
        self.action_name = action_name
        self.parameter_list = parameter_list
        
    @property
    def children(self):
        return (self.handle, self.parameter_list)
    

class FunctionInvocationNode(Node):
    action_name = None
    parameter_list = None
    
    def __init__(self, action_name, parameter_list):
        self.action_name = action_name
        self.parameter_list = parameter_list
        
    @property
    def children(self):
        return (self.parameter_list,)
    
    
class ImplicitInvocationNode(Node):
    namespace = None
    action_name = None
    parameter_list = None
    
    def __init__(self, namespace, action_name, parameter_list):
        self.namespace = namespace
        self.action_name = action_name
        self.parameter_list = parameter_list

    @property
    def children(self):
        return (self.parameter_list,)


class ClassInvocationNode(ImplicitInvocationNode):
    
    @property
    def key_letter(self):
        return self.namespace
    
    
class BridgeInvocationNode(ImplicitInvocationNode):
    
    @property
    def key_letter(self):
        return self.namespace
    
    
class PortInvocationNode(ImplicitInvocationNode):
    
    @property
    def port_name(self):
        return self.namespace


class GeneratePortEventNode(Node):
    port_name = None
    action_name = None
    parameter_list = None
    expression = None
    
    def __init__(self, port_name, action_name, parameter_list, expression):
        self.port_name = port_name
        self.action_name = action_name
        self.parameter_list = parameter_list
        self.expression = expression
        
    @property
    def children(self):
        return (self.parameter_list, self.expression)


class ParameterListNode(Node):
    children = None
    
    def __init__(self):
        self.children = list()
           

class ParameterNode(Node):
    name = None
    expression = None
    
    def __init__(self, name, expression):
        self.name = name
        self.expression = expression
        
    @property
    def children(self):
        return (self.expression,)


class UnaryOperationNode(Node):
    operator = None
    operand = None
    
    def __init__(self, operator, operand):
        self.operator = operator
        self.operand = operand
        
    @property
    def children(self):
        return (self.operand,)


class BinaryOperationNode(Node):
    left = None
    operator = None
    right = None
    
    def __init__(self, left, operator, right):
        self.left = left
        self.operator = operator
        self.right = right
        
    @property
    def children(self):
        return (self.left, self.right)


class VariableAccessNode(Node):
    variable_name = None
    
    def __init__(self, variable_name):
        self.variable_name = variable_name


class SelfAccessNode(VariableAccessNode):

    def __init__(self, variable_name='self'):
        VariableAccessNode.__init__(self, variable_name)


class SelectedAccessNode(VariableAccessNode):

    def __init__(self, variable_name='selected'):
        VariableAccessNode.__init__(self, variable_name)


class ParamAccessNode(Node):
    variable_name = None
    
    def __init__(self, variable_name):
        self.variable_name = variable_name
    

class FieldAccessNode(Node):
    handle = None
    name = None
    
    def __init__(self, handle, name):
        self.handle = handle
        self.name = name

    @property
    def children(self):
        return (self.handle,)


class IndexAccessNode(Node):
    handle = None
    expression = None
    
    def __init__(self, handle, expression):
        self.handle = handle
        self.expression = expression

    @property
    def children(self):
        return (self.handle, self.expression)
    

class IntegerNode(Node):
    value = None
    
    def __init__(self, value):
        self.value = value


class RealNode(Node):
    value = None
    
    def __init__(self, value):
        self.value = value


class StringNode(Node):
    value = None
    
    def __init__(self, value):
        self.value = value


class BooleanNode(Node):
    value = None
    
    def __init__(self, value):
        self.value = value


class EnumOrNamedConstantNode(Node):
    namespace = None
    name = None
    
    def __init__(self, namespace, name):
        self.namespace = namespace
        self.name = name


def find_column(lexdata, lexpos):
    '''
    Find the column index
    '''
    return lexpos - lexdata.rfind('\n', 0, lexpos)


def set_positional_info(node, p):
    '''
    set positional information on a node
    '''
    node.position = Position()
    node.position.label = p.lexer.label
    node.position.start_stream = p.lexpos(1)
    node.position.start_line = p.lineno(1)
    node.position.start_column = find_column(p.lexer.lexdata,
                                             node.position.start_stream)
    
    _, node.position.end_stream = p.lexspan(len(p) - 1)
    _, node.position.end_line = p.linespan(len(p) - 1)
    node.position.end_column = find_column(p.lexer.lexdata,
                                             node.position.end_stream) - 1
    
    node.character_stream = p.lexer.lexdata[node.position.start_stream:
                                            node.position.end_stream]
    
def track_production(f):
    '''
    decorator for adding positional information to returning nodes
    '''
    @wraps(f)
    def wrapper(self, p):
        r = f(self, p)
        node = p[0]
        if isinstance(node, Node) and len(p) > 1:
            set_positional_info(node, p)
        return r
    
    return wrapper


class ParseException(Exception):
    pass


class OALParser(object):
    keywords = ('ASSIGN',
                'ASSIGNER',
                'BREAK',
                'BRIDGE',
                'SEND',
                'CONTROL',
                'STOP',
                'CONTINUE',
                'CREATE',
                'EVENT',
                'INSTANCE',
                'OF',
                'OBJECT',
                'DELETE',
                'FOR',
                'EACH',
                'IN',
                'GENERATE',
                'IF',
                'ELIF',
                'ELSE',
                'RELATE',
                'TO',
                'ACROSS',
                'USING',
                'RETURN',
                'SELECT',
                'ONE',
                'ANY',
                'MANY',
                'TRANSFORM',
                'UNRELATE',
                'FROM',
                'WHILE',
                'CLASS',
                'CREATOR',
                'RELATED',
                'BY',
                'INSTANCES',
                'WHERE',
                'CARDINALITY',
                'EMPTY',
                'FALSE',
                'NOT',
                'NOT_EMPTY',
                'TRUE',
                'AND',
                'OR',
                'PARAM',
                'RCVD_EVT',
                'SELF',
                'SELECTED'
    )
    
    tokens = keywords + (
                         'SEMICOLON',
                         'EQUAL',
                         'DOT',
                         'DOUBLECOLON',
                         'LPAREN',
                         'RPAREN',
                         'TIMES',
                         'COLON',
                         'COMMA',
                         'ARROW',
                         'LSQBR',
                         'RSQBR',
                         'ID',
                         'NAMESPACE',
                         'END_FOR',
                         'END_IF',
                         'END_WHILE',
                         'TICKED_PHRASE',
                         'QMARK',
                         'FRACTION',
                         'NUMBER',
                         'STRING',
                         'DOUBLEEQUAL',
                         'NOTEQUAL',
                         'LESSTHAN',
                         'LE',
                         'GT',
                         'GE',
                         'PLUS',
                         'MINUS',
                         'PIPE',
                         'DIV',
                         'MOD',
                         'AMP',
                         'CARET',
                         'COMMENT',
                         'SL_STRING'
        )
    
    
    t_ignore = ' \t\r'
    
    precedence = (
        ('left', 'OR'),
        ('left', 'AND'),
        ('nonassoc', 'LESSTHAN', 'LE', 'DOUBLEEQUAL', 'GT', 'GE', 'NOTEQUAL'),
        ('left', 'PLUS', 'MINUS', 'PIPE'),
        ('left', 'TIMES', 'DIV', 'AMP', 'CARET'),
        ('left', 'MOD'),
        ('right', 'UNARY'),
    )
    
    def __init__(self):
        self.parser = yacc.yacc(debuglog=logger,
                                errorlog=logger,
                                optimize=1,
                                module=self,
                                outputdir=os.path.dirname(__file__),
                                tabmodule='bridgepoint.__oal_parsetab')

    def text_input(self, text, label='<string>'):
        lexer = lex.lex(debuglog=logger,
                        errorlog=logger,
                        optimize=1,
                        module=self,
                        outputdir=os.path.dirname(__file__),
                        lextab="bridgepoint.__oal_lextab")

        lexer.label = label
        return self.parser.parse(lexer=lexer,
                                 input=text,
                                 tracking=1)

    def t_COMMENT(self, t):
        r'/\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+/'
        t.lexer.lineno += t.value.count('\n')
        t.endlexpos = t.lexpos + len(t.value)
    
    def t_SL_STRING(self, t):
        r'\/\/.*\n'
        t.lexer.lineno += t.value.count('\n')
        t.endlexpos = t.lexpos + len(t.value)
    
    def t_TICKED_PHRASE(self, t):
        r"\'[^\']*\'"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_STRING(self, t):
        r'"[^"\n]*"'
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_END_FOR(self, t):
        r"(?i)end[\s]+for"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_END_IF(self, t):
        r"(?i)end[\s]+if"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_END_WHILE(self, t):
        r"(?i)end[\s]+while"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_NAMESPACE(self, t):
        r"([0-9a-zA-Z_])+(?=::)"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_ID(self, t):
        r"[a-zA-Z_][0-9a-zA-Z_]*|[a-zA-Z][0-9a-zA-Z_]*[0-9a-zA-Z_]+"
        t.endlexpos = t.lexpos + len(t.value)
        
        value = t.value.upper()
        if value in self.keywords:
            t.type = value
            
        return t
        
    def t_FRACTION(self, t):
        r"(((\d*\.\d+)|(\d+\.)([eE][-+]?\d+)?)|(\d+([eE][-+]?\d+)))[FfLl]?"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_NUMBER(self, t):
        r"\d+"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_DOUBLECOLON(self, t):
        r"::"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_DOUBLEEQUAL(self, t):
        r"\=\="
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_NOTEQUAL(self, t):
        r"!\="
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_ARROW(self, t):
        r"\-\>"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_LE(self, t):
        r"\<\="
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_GE(self, t):
        r"\>\="
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_SEMICOLON(self, t):
        r";"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_EQUAL(self, t):
        r"\="
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_DOT(self, t):
        r"\."
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_LPAREN(self, t):
        r"\("
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_RPAREN(self, t):
        r"\)"
        t.endlexpos = t.lexpos + len(t.value)
        return t
        
    def t_TIMES(self, t):
        r"\*"
        t.endlexpos = t.lexpos + len(t.value)
        return t
        
    def t_COLON(self, t):
        r":"
        t.endlexpos = t.lexpos + len(t.value)
        return t
        
    def t_COMMA(self, t):
        r","
        t.endlexpos = t.lexpos + len(t.value)
        return t
        
    def t_LSQBR(self, t):
        r"\["
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_RSQBR(self, t):
        r"\]"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_QMARK(self, t):
        r"\?"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_LESSTHAN(self, t):
        r"\<"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_GT(self, t):
        r"\>"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_PLUS(self, t):
        r"\+"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_MINUS(self, t):
        r"\-"
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_PIPE(self, t):
        r"\|"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_DIV(self, t):
        r"/"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_MOD(self, t):
        r"%"
        t.endlexpos = t.lexpos + len(t.value)
        return t
 
    def t_AMP(self, t):
        r"&"
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_CARET(self, t):
        r"\^"
        t.endlexpos = t.lexpos + len(t.value)
        return t
 
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)
        t.endlexpos = t.lexpos + len(t.value)
        
    def t_error(self, t):
        logger.error("%d,%d:illegal character '%s'in INITIAL" % (t.lineno, t.lexpos, t.value[0]))
        t.lexer.skip(1)
    
    @track_production
    def p_action(self, p):
        '''action : action_body'''
        p[0] = p[1]
         
    @track_production
    def p_action_body(self, p):
        '''action_body : block'''
        p[0] = BodyNode(block=p[1])
    
    @track_production
    def p_block(self, p):
        '''block : statement_list'''
        p[0] = BlockNode(statement_list=p[1])
    
    @track_production
    def p_empty_block(self, p):
        '''block : '''
        p[0] = BlockNode(statement_list=StatementListNode())
        
    @track_production       
    def p_statement_list_1(self, p):
        '''statement_list : statement SEMICOLON statement_list'''
        p[0] = p[3]
        if p[1] is not None:
            p[0].children.insert(0, p[1])
    
    @track_production
    def p_statement_list_2(self, p):
        '''statement_list : statement SEMICOLON'''
        p[0] = StatementListNode()
        if p[1] is not None:
            p[0].children.insert(0, p[1])
    
    @track_production
    def p_empty_statement(self, p):
        '''statement : '''
        pass
    
    @track_production
    def p_break_statement(self, p):
        '''statement : BREAK'''
        p[0] = BreakNode()
    
    @track_production
    def p_continue_statement(self, p):
        '''statement : CONTINUE'''
        p[0] = ContinueNode()
    
    @track_production
    def p_control_statement(self, p):
        '''statement : CONTROL STOP'''
        p[0] = ControlNode()
        
    @track_production
    def p_return_statement_1(self, p):
        '''statement : RETURN expression'''
        p[0] = ReturnNode(expression=p[2])
    
    @track_production
    def p_return_statement_2(self, p):
        '''statement : RETURN'''
        p[0] = ReturnNode(expression=None)

    @track_production
    def p_explicit_assignment_statement(self, p):
        '''statement : ASSIGN variable_access EQUAL expression'''
        p[0] = AssignmentNode(variable_access=p[2],
                              expression=p[4])
    
    @track_production
    def p_implicit_assignment_statement(self, p):
        '''statement : variable_access EQUAL expression'''
        p[0] = AssignmentNode(variable_access=p[1],
                              expression=p[3])
    
    @track_production
    def p_function_invocation_statement(self, p):
        '''statement : invocation'''
        p[0] = InvocationStatementNode(p[1])
    
    @track_production
    def p_bridge_assignment_statement(self, p):
        '''statement : BRIDGE variable_access EQUAL implicit_invocation'''
        p[4].__class__ = BridgeInvocationNode
        p[0] = AssignmentNode(variable_access=p[2],
                              expression=p[4])
        
    @track_production
    def p_bridge_invocation_statement(self, p):
        '''statement : BRIDGE implicit_invocation'''
        p[2].__class__ = BridgeInvocationNode
        p[0] = InvocationStatementNode(p[2])
        
    @track_production
    def p_instance_invocation_statement(self, p):
        '''statement : TRANSFORM instance_invocation'''
        p[0] = InvocationStatementNode(p[2])
    
    @track_production
    def p_instance_invocation_assignment_statement(self, p):
        '''statement : TRANSFORM variable_access EQUAL instance_invocation'''
        p[0] = AssignmentNode(variable_access=p[2],
                              expression=p[4])
        
    @track_production
    def p_class_invocation_statement(self, p):
        '''statement : TRANSFORM implicit_invocation'''
        p[2].__class__ = ClassInvocationNode
        p[0] = InvocationStatementNode(p[2])
        
    @track_production
    def p_class_invocation_assignment_statement(self, p):
        '''statement : TRANSFORM variable_access EQUAL implicit_invocation'''
        p[4].__class__ = ClassInvocationNode
        p[0] = AssignmentNode(variable_access=p[2],
                              expression=p[4])
    
    @track_production
    def p_port_invocation_statement(self, p):
        '''statement : SEND implicit_invocation'''
        p[2].__class__ = PortInvocationNode
        p[0] = InvocationStatementNode(p[2])
        
    @track_production
    def p_port_invocation_assignment_statement(self, p):
        '''statement : SEND variable_access EQUAL implicit_invocation'''
        p[4].__class__ = PortInvocationNode
        p[0] = AssignmentNode(variable_access=p[2],
                              expression=p[4])
    
    @track_production
    def p_port_event_generation(self, p):
        '''statement : SEND namespace DOUBLECOLON identifier LPAREN parameter_list RPAREN TO expression'''
        p[0] = GeneratePortEventNode(port_name=p[2],
                                     action_name=p[4],
                                     parameter_list=p[6],
                                     expression=p[9])
        
    @track_production
    def p_generate_class_event_statement(self, p):
        '''statement : GENERATE event_specification TO identifier CLASS'''
        p[0] = GenerateClassEventNode(event_specification=p[2],
                                      key_letter=p[4])
    
    @track_production
    def p_generate_assigner_event_statement(self, p):
        '''statement : GENERATE event_specification TO identifier ASSIGNER'''
        p[0] = GenerateClassEventNode(event_specification=p[2],
                                      key_letter=p[4])
        
    @track_production
    def p_generate_creator_event_statement(self, p):
        '''statement : GENERATE event_specification TO identifier CREATOR'''
        p[0] = GenerateCreatorEventNode(event_specification=p[2],
                                        key_letter=p[4])
    
    @track_production
    def p_generate_instance_event_statement_1(self, p):
        '''statement : GENERATE event_specification TO variable_access'''
        p[0] = GenerateInstanceEventNode(event_specification=p[2],
                                         variable_access=p[4])
        
    @track_production
    def p_generate_instance_event_statement_2(self, p):
        '''statement : GENERATE event_specification TO self_access'''
        p[0] = GenerateInstanceEventNode(event_specification=p[2],
                                         variable_access=p[4])
        
    @track_production
    def p_create_class_event_statement(self, p):
        '''statement : CREATE EVENT INSTANCE variable_name OF event_specification TO identifier CLASS'''
        p[0] = CreateClassEventNode(variable_name=p[4],
                                    event_specification=p[6],
                                    key_letter=p[8])
        
    @track_production
    def p_create_assigner_event_statement(self, p):
        '''statement : CREATE EVENT INSTANCE variable_name OF event_specification TO identifier ASSIGNER'''
        p[0] = CreateClassEventNode(variable_name=p[4],
                                    event_specification=p[6],
                                    key_letter=p[8])
        
    @track_production
    def p_create_creator_event_statement(self, p):
        '''statement : CREATE EVENT INSTANCE variable_name OF event_specification TO identifier CREATOR'''
        p[0] = CreateCreatorEventNode(variable_name=p[4],
                                      event_specification=p[6],
                                      key_letter=p[8])
    
    @track_production
    def p_create_instance_event_statement_1(self, p):
        '''statement : CREATE EVENT INSTANCE variable_name OF event_specification TO variable_access'''
        p[0] = CreateInstanceEventNode(variable_name=p[4],
                                       event_specification=p[6],
                                       to_variable_access=p[8])
    
    @track_production
    def p_create_instance_event_statement_2(self, p):
        '''statement : CREATE EVENT INSTANCE variable_name OF event_specification TO self_access'''
        p[0] = CreateInstanceEventNode(variable_name=p[4],
                                       event_specification=p[6],
                                       to_variable_access=p[8])
        
    @track_production
    def p_generate_preexisting_event_statement(self, p):
        '''statement : GENERATE variable_access'''
        p[0] = GeneratePreexistingNode(variable_access=p[2])
    
    @track_production
    def p_event_specification(self, p):
        '''event_specification : identifier event_meaning event_data'''
        p[0] = EventSpecNode(identifier=p[1],
                             meaning=p[2],
                             event_data=p[3])
    
    @track_production
    def p_ploymorphic_event_spec(self, p):
        '''event_specification : identifier TIMES event_meaning event_data'''
        p[0] = EventSpecNode(identifier=p[1],
                             meaning=p[3],
                             event_data=p[4])
    
    
    def p_phrase_1(self, p):
        '''phrase : TICKED_PHRASE'''
        p[0] = p[1]
        
    def p_phrase_2(self, p):
        '''phrase : identifier'''
        p[0] = "'%s'" % p[1]
        
    @track_production
    def p_event_meaning(self, p):
        '''event_meaning : COLON phrase'''
        p[0] = p[2]
            
    @track_production
    def p_event_meaning_4(self, p):
        '''event_meaning : '''
        p[0] = None
    
    @track_production
    def p_event_data_1(self, p):
        '''event_data : LPAREN event_parameter_list RPAREN'''
        p[0] = p[2]
    
    @track_production
    def p_event_data_2(self, p):
        '''event_data : '''
        p[0] = EventDataListNode()

    @track_production
    def p_event_parameter_list_1(self, p):
        '''event_parameter_list : event_parameter'''
        p[0] = EventDataListNode()
        p[0].children.append(p[1])
    
    @track_production
    def p_event_parameter_list_2(self, p):
        '''event_parameter_list : event_parameter COMMA event_parameter_list'''
        p[0] = p[3]
        p[0].children.insert(0, p[1])
    
    @track_production
    def p_event_parameter_list_3(self, p):
        '''event_parameter_list : '''
        p[0] = EventDataListNode()
    
    @track_production
    def p_event_parameter(self, p):
        '''event_parameter : identifier COLON expression'''
        p[0] = EventDataItemNode(name=p[1],
                                 expression=p[3])

    @track_production
    def p_create_object_statement(self, p):
        '''statement : CREATE OBJECT INSTANCE variable_name OF identifier'''
        p[0] = CreateObjectNode(variable_name=p[4],
                                key_letter=p[6])
    
    @track_production
    def p_create_object_no_variable_statement(self, p):
        '''statement : CREATE OBJECT INSTANCE OF identifier'''
        p[0] = CreateObjectNoVariableNode(key_letter=p[5])
        
    @track_production
    def p_delete_statement(self, p):
        '''statement : DELETE OBJECT INSTANCE instance_name'''
        p[0] = DeleteNode(variable_name=p[4])
    
    @track_production
    def p_for_statement(self, p):
        '''statement : FOR EACH variable_name IN variable_name block END_FOR'''
        p[0] = ForEachNode(instance_variable_name=p[3],
                           set_variable_name=p[5],
                           block=p[6])
    
    @track_production
    def p_while_statement(self, p):
        '''statement : WHILE expression block END_WHILE'''
        p[0] = WhileNode(expression=p[2],
                         block=p[3])
    
    @track_production
    def p_if_statement(self, p):
        '''statement : IF expression block elif_list else_clause END_IF'''
        p[0] = IfNode(expression=p[2],
                      block=p[3],
                      elif_list=p[4],
                      else_clause=p[5])
    
    @track_production       
    def p_elif_list_1(self, p):
        '''elif_list : '''
        p[0] = ElIfListNode()
            
    @track_production       
    def p_elif_list_2(self, p):
        '''elif_list : ELIF elif_clause elif_list'''
        p[0] = p[3]
        p[0].children.insert(0, p[2])
        
    @track_production
    def p_elif_clause(self, p):
        '''elif_clause : expression block'''
        p[0] = ElIfNode(expression=p[1],
                        block=p[2])

    @track_production
    def p_else_clause_1(self, p):
        '''else_clause : ELSE block'''
        p[0] = ElseNode(p[2])
    
    @track_production
    def p_else_clause_2(self, p):
        '''else_clause : '''
        p[0] = None
    
    def p_rel_id(self, p):
        '''rel_id : limited_identifier'''
        p[0] = p[1]
        
    @track_production
    def p_relate_statement_1(self, p):
        '''statement : RELATE instance_name TO instance_name ACROSS rel_id'''
        p[0] = RelateNode(from_variable_name=p[2],
                          to_variable_name=p[4],
                          rel_id=p[6],
                          phrase=None)
    
    @track_production
    def p_relate_statement_2(self, p):
        '''statement : RELATE instance_name TO instance_name ACROSS rel_id DOT phrase'''
        p[0] = RelateNode(from_variable_name=p[2],
                          to_variable_name=p[4],
                          rel_id=p[6],
                          phrase=p[8])
    
    @track_production
    def p_relate_using_statement_1(self, p):
        '''statement : RELATE instance_name TO instance_name ACROSS rel_id USING instance_name'''
        p[0] = RelateUsingNode(from_variable_name=p[2],
                               to_variable_name=p[4],
                               rel_id=p[6],
                               phrase=None,
                               using_variable_name=p[8])
    
    @track_production
    def p_relate_using_statement_2(self, p):
        '''statement : RELATE instance_name TO instance_name ACROSS rel_id DOT phrase USING instance_name'''
        p[0] = RelateUsingNode(from_variable_name=p[2],
                               to_variable_name=p[4],
                               rel_id=p[6],
                               phrase=p[8],
                               using_variable_name=p[10])
    
    @track_production
    def p_unrelate_statement_1(self, p):
        '''statement : UNRELATE instance_name FROM instance_name ACROSS rel_id'''
        p[0] = UnrelateNode(from_variable_name=p[2],
                            to_variable_name=p[4],
                            rel_id=p[6],
                            phrase=None)
    
    @track_production
    def p_unrelate_statement_2(self, p):
        '''statement : UNRELATE instance_name FROM instance_name ACROSS rel_id DOT phrase'''
        p[0] = UnrelateNode(from_variable_name=p[2],
                            to_variable_name=p[4],
                            rel_id=p[6],
                            phrase=p[8])
    
    @track_production
    def p_unrelate_statement_using_1(self, p):
        '''statement : UNRELATE instance_name FROM instance_name ACROSS rel_id USING instance_name'''
        p[0] = UnrelateUsingNode(from_variable_name=p[2],
                                 to_variable_name=p[4],
                                 rel_id=p[6],
                                 phrase=None,
                                 using_variable_name=p[8])
    
    @track_production
    def p_unrelate_statement_using_2(self, p):
        '''statement : UNRELATE instance_name FROM instance_name ACROSS rel_id DOT phrase USING instance_name'''
        p[0] = UnrelateUsingNode(from_variable_name=p[2],
                                 to_variable_name=p[4],
                                 rel_id=p[6],
                                 phrase=p[8],
                                 using_variable_name=p[10])
    
    @track_production
    def p_select_from_statement_1(self, p):
        '''
        statement : SELECT ANY variable_name FROM INSTANCES OF identifier
                  | SELECT MANY variable_name FROM INSTANCES OF identifier
        '''
        p[0] = SelectFromNode(cardinality=p[2],
                              variable_name=p[3],
                              key_letter=p[7])
                              
    @track_production
    def p_select_from_statement_2(self, p):
        '''
        statement : SELECT ANY variable_name FROM identifier
                  | SELECT MANY variable_name FROM identifier
        '''
        p[0] = SelectFromNode(cardinality=p[2],
                              variable_name=p[3],
                              key_letter=p[5])
    
    @track_production
    def p_select_from_where_statement_1(self, p):
        '''
        statement : SELECT ANY variable_name FROM INSTANCES OF identifier WHERE expression
                  | SELECT MANY variable_name FROM INSTANCES OF identifier WHERE expression
        '''
        p[0] = SelectFromWhereNode(cardinality=p[2],
                                   variable_name=p[3],
                                   key_letter=p[7],
                                   where_clause=p[9])
    
    @track_production
    def p_select_from_where_statement_2(self, p):
        '''
        statement : SELECT ANY variable_name FROM identifier WHERE expression
                  | SELECT MANY variable_name FROM identifier WHERE expression
        '''
        p[0] = SelectFromWhereNode(cardinality=p[2],
                                   variable_name=p[3],
                                   key_letter=p[5],
                                   where_clause=p[7])
                                   
    def p_navigation_hook(self, p):
        '''
        navigation_hook : variable_access
                        | self_access
        '''
        p[0] = p[1]
        
    @track_production
    def p_select_related_statement(self, p):
        '''
        statement : SELECT ONE  variable_name RELATED BY navigation_hook navigation_chain
                  | SELECT ANY  variable_name RELATED BY navigation_hook navigation_chain
                  | SELECT MANY variable_name RELATED BY navigation_hook navigation_chain
        '''
        p[0] = SelectRelatedNode(cardinality=p[2],
                                 variable_name=p[3],
                                 handle=p[6],
                                 navigation_chain=p[7])
    
    @track_production
    def p_select_related_where_statement(self, p):
        '''
        statement : SELECT ONE  variable_name RELATED BY navigation_hook navigation_chain WHERE expression
                  | SELECT ANY  variable_name RELATED BY navigation_hook navigation_chain WHERE expression
                  | SELECT MANY variable_name RELATED BY navigation_hook navigation_chain WHERE expression
        '''
        p[0] = SelectRelatedWhereNode(cardinality=p[2],
                                      variable_name=p[3],
                                      handle=p[6],
                                      navigation_chain=p[7],
                                      where_clause=p[9])
        
    @track_production
    def p_navigation_chain_1(self, p):
        '''navigation_chain : navigation_step'''
        p[0] = NavigationListNode()
        p[0].children.append(p[1])
    
    @track_production
    def p_navigation_chain_2(self, p):
        '''navigation_chain : navigation_step navigation_chain'''
        p[0] = p[2]
        p[0].children.insert(0, p[1])
    
    @track_production
    def p_navigation_step_1(self, p):
        '''navigation_step : ARROW identifier LSQBR identifier RSQBR'''
        p[0] = NavigationStepNode(key_letter=p[2],
                                  rel_id=p[4],
                                  phrase=None)

    @track_production
    def p_navigation_step_2(self, p):
        '''navigation_step : ARROW identifier LSQBR identifier DOT phrase RSQBR'''
        p[0] = NavigationStepNode(key_letter=p[2],
                                  rel_id=p[4],
                                  phrase=p[6])

    @track_production
    def p_invocation(self, p):
        '''
        invocation : implicit_invocation
                   | function_invocation
                   | instance_invocation
        '''
        p[0] = p[1]
        
    @track_production
    def p_implicit_invocation(self, p):
        '''implicit_invocation : namespace DOUBLECOLON identifier LPAREN parameter_list RPAREN'''
        p[0] = ImplicitInvocationNode(namespace=p[1],
                                      action_name=p[3],
                                      parameter_list=p[5])
    
    @track_production
    def p_function_invocation(self, p):
        '''function_invocation : DOUBLECOLON identifier LPAREN parameter_list RPAREN'''
        p[0] = FunctionInvocationNode(action_name=p[2],
                                      parameter_list=p[4])
    
    @track_production
    def p_operation_invocation_1(self, p):
        '''instance_invocation : structure DOT identifier LPAREN parameter_list RPAREN'''
        p[0] = InstanceInvocationNode(handle=p[1],
                                      action_name=p[3],
                                      parameter_list=p[5])

    @track_production
    def p_parameter_list_1(self, p):
        '''parameter_list : parameter'''
        p[0] = ParameterListNode()
        p[0].children.append(p[1])
    
    @track_production
    def p_parameter_list_2(self, p):
        '''parameter_list : parameter COMMA parameter_list'''
        p[0] = p[3]
        p[0].children.insert(0, p[1])
    
    @track_production
    def p_parameter_list_3(self, p):
        '''parameter_list : '''
        p[0] = ParameterListNode()
    
    @track_production
    def p_parameter(self, p):
        '''parameter : identifier COLON expression'''
        p[0] = ParameterNode(name=p[1],
                             expression=p[3])
        
    @track_production
    def p_expression_1(self, p):
        '''expression : constant'''
        p[0] = p[1]
    
    @track_production
    def p_expression_2(self, p):
        '''expression : variable_access'''
        p[0] = p[1]

    @track_production
    def p_expression_4(self, p):
        '''expression : self_access'''
        p[0] = p[1]
        
    @track_production
    def p_expression_5(self, p):
        '''expression : selected_access'''
        p[0] = p[1]
        
    @track_production
    def p_grouped_expression(self, p):
        '''expression : LPAREN expression RPAREN'''
        p[0] = p[2]   

    @track_production
    def p_invocation_expression(self, p):
        '''expression : invocation'''
        p[0] = p[1]

    @track_production
    def p_unary_expression(self, p):
        '''expression : unary_operator expression %prec UNARY'''
        p[0] = UnaryOperationNode(operator=p[1],
                                  operand=p[2])
    
    @track_production
    def p_arithmetic_expression(self, p):
        '''
        expression : expression PLUS  expression
                   | expression MINUS expression
                   | expression PIPE  expression
                   | expression TIMES expression
                   | expression DIV   expression
                   | expression MOD   expression
                   | expression AMP   expression
                   | expression CARET expression
        '''
        p[0] = BinaryOperationNode(left=p[1],
                                   operator=p[2],
                                   right=p[3])
    
    @track_production
    def p_boolean_expression(self, p):
        '''
        expression : expression LE          expression
                   | expression LESSTHAN    expression
                   | expression DOUBLEEQUAL expression
                   | expression NOTEQUAL    expression
                   | expression GE          expression
                   | expression GT          expression
                   | expression AND         expression
                   | expression OR          expression
        '''
        p[0] = BinaryOperationNode(left=p[1],
                                   operator=p[2],
                                   right=p[3])
    
    @track_production
    def p_limited_identifier(self, p):
        '''
        limited_identifier : ID
                           | kw_as_identifier_1
        '''
        p[0] = p[1]
        
    @track_production
    def p_namespace(self, p):
        '''
        namespace : NAMESPACE
        '''
        p[0] = p[1]
    
    @track_production
    def p_kw_as_identifier_1(self, p):
        '''kw_as_identifier_1 : ACROSS
                              | ANY
                              | ASSIGN
                              | ASSIGNER
                              | BREAK
                              | BY
                              | CLASS
                              | CONTINUE
                              | CONTROL
                              | CREATE
                              | CREATOR
                              | DELETE
                              | EACH
                              | EVENT
                              | FOR
                              | FROM
                              | GENERATE
                              | IN
                              | INSTANCES
                              | INSTANCE
                              | MANY
                              | OBJECT
                              | ONE
                              | RELATED
                              | RELATE
                              | SELECT
                              | STOP
                              | TO
                              | WHERE
                              | UNRELATE
                              | USING
        '''
        p[0] = p[1]
        #FIXME: the oal parser currently does not allow variables named 'of'
    
    @track_production
    def p_kw_as_identifier_2(self, p):
        '''
        kw_as_identifier_2 : BRIDGE
                           | CARDINALITY
                           | EMPTY
                           | FALSE
                           | NOT
                           | NOT_EMPTY
                           | SEND
                           | TRANSFORM
                           | TRUE
                           | OF
        '''
        p[0] = p[1]
        
    @track_production
    def p_kw_as_identifier_3(self, p):
        '''
        kw_as_identifier_3 : PARAM
                           | RCVD_EVT
                           | SELECTED
                           | SELF
        '''
        p[0] = p[1]
        
    @track_production
    def p_kw_as_identifier_4(self, p):
        '''
        kw_as_identifier_4 : AND
                           | ELIF
                           | ELSE
                           | IF
                           | OR
                           | RETURN
                           | WHILE
        '''
        p[0] = p[1]
                
    @track_production
    def p_variable_name(self, p):
        '''variable_name : limited_identifier'''
        p[0] = p[1]

    @track_production
    def p_instance_name(self, p):
        '''instance_name : variable_name
                         | SELF
        '''
        p[0] = p[1]
        
    @track_production
    def p_identifier(self, p):
        '''
        identifier : limited_identifier
              | kw_as_identifier_2
              | kw_as_identifier_3
              | kw_as_identifier_4
        '''
        p[0] = p[1]
        
    @track_production
    def p_structure_1(self, p):
        '''structure : variable_name'''
        p[0] = VariableAccessNode(variable_name=p[1])
        
    @track_production
    def p_structure_2(self, p):
        '''structure : SELF'''
        p[0] = SelfAccessNode()
        
    @track_production
    def p_structure_3(self, p):
        '''structure : SELECTED'''
        p[0] = SelectedAccessNode()
        
    @track_production
    def p_field_access(self, p):
        '''
        field_access : structure    DOT identifier
                     | index_access DOT identifier
                     | field_access DOT identifier
                     | param_access DOT identifier
        '''
        p[0] = FieldAccessNode(handle=p[1], name=p[3])

    @track_production
    def p_array(self, p):
        '''array : variable_name'''
        p[0] = VariableAccessNode(variable_name=p[1])
        
    @track_production
    def p_index_access(self, p):
        '''index_access : array        LSQBR expression RSQBR
                        | index_access LSQBR expression RSQBR
                        | field_access LSQBR expression RSQBR
                        | param_access LSQBR expression RSQBR
        '''
        p[0] = IndexAccessNode(handle=p[1], expression=p[3])
        
    @track_production
    def p_param(self, p):
        '''
        param : PARAM
              | RCVD_EVT
        '''
        p[0] = p[1]
        
    @track_production
    def p_param_access_1(self, p):
        '''param_access : param DOT variable_name'''
        p[0] = ParamAccessNode(variable_name=p[3])

    @track_production
    def p_self_access_1(self, p):
        '''self_access : SELF'''
        p[0] = SelfAccessNode()
        
    @track_production
    def p_selected_access_1(self, p):
        '''selected_access : SELECTED'''
        p[0] = SelectedAccessNode()
    
    @track_production
    def p_variable_access_1(self, p):
        '''variable_access : variable_name'''
        p[0] = VariableAccessNode(variable_name=p[1])
        
    @track_production
    def p_variable_access_2(self, p):
        '''variable_access : field_access'''
        p[0] = p[1]
        
    @track_production
    def p_variable_access_3(self, p):
        '''variable_access : index_access'''
        p[0] = p[1]
        
    @track_production
    def p_variable_access_4(self, p):
        '''variable_access : param_access'''
        p[0] = p[1]
        
    @track_production
    def p_constant_fraction(self, p):
        '''constant : FRACTION'''
        p[0] = RealNode(value=p[1])

    @track_production
    def p_constant_number(self, p):
        '''constant : NUMBER'''
        p[0] = IntegerNode(value=p[1])
    
    @track_production
    def p_constant_string(self, p):
        '''constant : STRING'''
        p[0] = StringNode(value=p[1])
    
    @track_production
    def p_constant_boolean(self, p):
        '''
        constant : TRUE
                 | FALSE
        '''
        p[0] = BooleanNode(value=p[1])
    
    @track_production
    def p_enum_or_named_constant(self, p):
        '''constant : namespace DOUBLECOLON identifier'''
        p[0] = EnumOrNamedConstantNode(namespace=p[1],
                                       name=p[3])

    @track_production
    def p_unary_operator(self, p):
        '''
        unary_operator : NOT
                       | EMPTY
                       | NOT_EMPTY
                       | CARDINALITY
                       | PLUS
                       | MINUS
        '''
        p[0] = p[1]
    
    def p_error(self, p):
        if p:
            raise ParseException("invalid token '%s' at line: %s, column: %s" % (p.type,p.lineno, find_column(p.lexer.lexdata, p.lexpos)))
        else:
            raise ParseException("unknown parsing error")


def parse(action_code, label='<string>'):
    '''
    Parse and construct an abstract syntax tree for text expressed in the
    Object Action Language (OAL).
    '''
    parser = OALParser()
    return parser.text_input(action_code + '\n', label)


if __name__ == '__main__':
    import sys
    import xtuml.tools
    logging.basicConfig(level=logging.WARN)
    
    print ('Enter the character stream below. Press Ctrl-D to begin parsing.')
    print ('')
    s = sys.stdin.read()
    
    print ('--------- Token Stream ----------')
    parser = OALParser()
    lexer = lex.lex(module=parser)
    lexer.input(s)
    while True:
        tok = lexer.token()
        if not tok:
            break
        print(tok)

    print ('--------- Syntax Tree ----------')
    root = parse(s)
    w = xtuml.tools.Walker()
    w.visitors.append(xtuml.tools.NodePrintVisitor())
    w.accept(root)


    



# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.

from bridgepoint import imp
imp.install()

from .prebuild import prebuild_action
from .prebuild import prebuild_model

from .sourcegen import gen_text_action

from .ooaofooa import ModelLoader
from .ooaofooa import load_metamodel
from .ooaofooa import load_component


#
# Suppress false import warning when invoking the command line interfaces:
#    python3 -m bridgepoint.consistency_check
#    python3 -m bridgepoint.interpret
#    python3 -m bridgepoint.oal
#    python3 -m bridgepoint.prebuild
#
# For details, see http://bugs.python.org/issue27487
#
import warnings
import sys

if not sys.warnoptions: # Overridable with -W flag
    warnings.filterwarnings('ignore', category=RuntimeWarning, module='runpy')




#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
Generate an sql schema file for an xtUML model. 
The arguments are either xtuml files, or folders containing *.xtuml files.
Note that some type of attributes are not supported, e.g. instance handles or
timers.
'''

import sys
import optparse
import logging

import xtuml

from . import ooaofooa


logger = logging.getLogger('gen_sql_schema')


def main():
    '''
    Parse argv for options and arguments, and start schema generation.
    '''
    parser = optparse.OptionParser(usage="%prog [options] <model_path> [another_model_path...]",
                                   version=xtuml.version.complete_string,
                                   formatter=optparse.TitledHelpFormatter())
                                   
    parser.set_description(__doc__.strip())
    
    parser.add_option("-c", "--component", dest="component", metavar="NAME",
                      help="export sql schema for the component named NAME",
                      action="store", default=None)
    
    parser.add_option("-d", "--derived-attributes", dest="derived",
                      help="include derived attributes in the schema",
                      action="store_true", default=False)
    
    parser.add_option("-o", "--output", dest='output', metavar="PATH",
                      help="save sql schema to PATH (required)",
                      action="store", default=None)
    
    parser.add_option("-v", "--verbosity", dest='verbosity', action="count", 
                      help="increase debug logging level", default=2)

    
    (opts, args) = parser.parse_args()
    if len(args) == 0 or opts.output is None:
        parser.print_help()
        sys.exit(1)

    levels = {
              0: logging.ERROR,
              1: logging.WARNING,
              2: logging.INFO,
              3: logging.DEBUG,
    }
    logging.basicConfig(level=levels.get(opts.verbosity, logging.DEBUG))

    loader = ooaofooa.Loader()
    for filename in args:
        loader.filename_input(filename)

    c = loader.build_component(opts.component, opts.derived)
    xtuml.persist_database(c, opts.output)

    
if __name__ == '__main__':
    main()




# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path

from bridgepoint import ooaofooa


def _get_version(filename):
    with open(filename, 'r') as f:
        line = f.readline()
        _, value = line.split("persistence-version: ", 1)
        return  value.strip()


def _check_magic(filename):
    if not os.path.exists(filename):
        return False
    
    with open(filename, 'r') as f:
        line = f.readline()
        return 'content: SystemModel' in line


class ModelImporter(object):
    
    def __init__(self, path):
        self.path = path
        
    @classmethod
    def find_module(cls, fullname, paths=None):
        if paths is None:
            paths = sys.path
                        
        names = fullname.split('.')
        model_name = names[-1]
        
        for f in paths:
            path = os.path.join(os.path.realpath(f), 'models')
            path = os.path.join(path, model_name)
            source = os.path.join(path, model_name + '.xtuml')
            if _check_magic(source):
                return cls(path)
    
    def get_code(self, module):
        pass

    def get_data(self, module):
        pass

    def get_filename(self, name):
        pass

    def get_source(self, name):
        pass

    def is_package(self, *args, **kwargs):
        return False

    def load_module(self, fullname):
        l = ooaofooa.Loader()
        l.filename_input(self.path)
        return l


def install():
    if ModelImporter not in sys.meta_path:
        sys.meta_path.append(ModelImporter)


def remove():
    sys.meta_path.remove(ModelImporter)




# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
Perform various xtUML meta operations, e.g. create new metamodels and
metaclasses, relate instances and perform navigations and queries.
'''

import logging

import xtuml

from functools import partial

try:
    from future_builtins import filter, zip
except ImportError:
    pass

import collections
try:
    # Python 3.x
    collectionsAbc = collections.abc
except:
    # Python 2.7
    collectionsAbc = collections


logger = logging.getLogger(__name__)


class MetaException(Exception):
    '''
    Base class for all exceptions thrown by the xtuml.meta module.
    '''


class DeleteException(MetaException):
    '''
    An exception that may be thrown during delete operations.
    '''


class RelateException(MetaException):
    '''
    An exception that may be thrown during relate operations.
    '''
    def __init__(self, from_instance, to_instance, rel_id, phrase):
        msg = '%s or %s already related across %s' % (from_instance,
                                                      to_instance,
                                                      rel_id)
        MetaException.__init__(self, msg)
    

class UnrelateException(MetaException):
    '''
    An exception that may be thrown during unrelate operations.
    '''
    def __init__(self, from_instance, to_instance, rel_id, phrase):
        msg = '%s and %s not related across %s' % (from_instance,
                                                   to_instance,
                                                   rel_id)
        MetaException.__init__(self, msg)


class UnknownLinkException(MetaException):
    '''
    An exception that may be thrown when a link is not found.
    '''
    def __init__(self, from_kind, to_kind, rel_id, phrase):
        if phrase:
            msg = "%s->%s[%s, %s]" % (from_kind, to_kind, repr(rel_id), repr(phrase))
        else:
            msg = "%s->%s[%s]" % (from_kind, to_kind, repr(rel_id))

        MetaException.__init__(self, 'Unknown link ' + msg)


class MetaModelException(MetaException):
    '''
    Base class for exceptions thrown by the MetaModel class.
    '''


class UnknownClassException(MetaModelException):
    '''
    An exception that may be thrown when a metaclass is not found.
    '''
    def __init__(self, kind):
        MetaModelException.__init__(self, 'Unknown class %s' % kind)


def _is_null(instance, name):
    '''
    Determine if an attribute of an *instance* with a specific *name* 
    is null.
    '''
    if name in instance.__dict__:
        value = instance.__dict__[name]
    else:
        value = getattr(instance, name)

    if value:
        return False
    
    elif value is None:
        return True

    name = name.upper()
    metaclass = get_metaclass(instance)
    for attr_name, attr_ty in metaclass.attributes:
        if attr_name.upper() != name:
            continue

        attr_ty = attr_ty.upper()
        if attr_ty == 'UNIQUE_ID':
            # UUID(int=0) is reserved for null
            return value == 0

        elif attr_ty == 'STRING':
            # empty string is reserved for null
            return len(value) == 0

        else:
            #null-values for integer, boolean and real are not supported
            return False


def apply_query_operators(iterable, ops):
    '''
    Apply a series of query operators to a sequence of instances, e.g.
    where_eq(), order_by() or filter functions.
    '''
    for op in ops:
        if isinstance(op, WhereEqual):
            iterable = op(iterable)
            
        elif isinstance(op, OrderBy):
            iterable = op(iterable)
            
        elif isinstance(op, dict):
            iterable = WhereEqual(op)(iterable)
            
        else:
            iterable = filter(op, iterable)

    return iterable


class Association(object):
    '''
    An association connects two metaclasses to each other via two directed
    links.
    '''
    rel_id = None
    source_keys = None
    source_link = None
    target_keys = None
    target_link = None
    
    def __init__(self, rel_id,
                 source_keys, source_link, 
                 target_keys, target_link):
        self.rel_id = rel_id
        self.source_link = source_link
        self.target_link = target_link
        self.source_keys = source_keys
        self.target_keys = target_keys
        
    @property
    def is_reflexive(self):
        return self.source_link.kind == self.target_link.kind

    def batch_relate(self):
        source_class = self.source_link.to_metaclass
        target_class = self.target_link.to_metaclass
        key_map = tuple(zip(self.source_keys, self.target_keys))
        
        for inst1 in source_class.storage:
            kwargs = dict()
            skip_instance = False
            
            for ref_key, primary_key in key_map:
                if _is_null(inst1, ref_key):
                    skip_instance = True
                    break
                
                kwargs[primary_key] = getattr(inst1, ref_key)
            
            if skip_instance:
                continue
            
            for inst2 in target_class.query(kwargs):
                self.source_link.connect(inst2, inst1, check=False)
                self.target_link.connect(inst1, inst2, check=False)

    def formalize(self):
        '''
        Formalize the association and expose referential attributes
        on instances.
        '''
        source_class = self.source_link.to_metaclass
        target_class = self.target_link.to_metaclass
        
        source_class.referential_attributes |= set(self.source_keys)
        target_class.identifying_attributes |= set(self.target_keys)

        def fget(inst, ref_name, alt_prop):
            other_inst = self.target_link.navigate_one(inst)
            if other_inst is None and alt_prop:
                return alt_prop.fget(inst)
            
            return getattr(other_inst, ref_name, None)

        def fset(inst, value, name, ref_name, alt_prop):
            kind = get_metaclass(inst).kind
            raise MetaException('%s.%s is a referential attribute '\
                                'and cannot be assigned directly'% (kind, name))
        
            #other_inst = self.target_link.navigate_one(inst)
            #if other_inst is None and alt_prop:
            #    return alt_prop.fset(inst, value)
            #
            #elif other_inst:
            #    return setattr(other_inst, ref_name, value)
        
        for ref_key, primary_key in zip(self.source_keys, self.target_keys):
            prop = getattr(source_class.clazz, ref_key, None)
            prop = property(partial(fget, ref_name=primary_key, alt_prop=prop), 
                            partial(fset, name=ref_key, ref_name=primary_key, alt_prop=prop))
            setattr(source_class.clazz, ref_key, prop)
            

class Link(dict):
    '''
    A link connects one metaclass to another in a single direction. 
    
    *rel_id* is used to specify the association between the to metaclasses,
    and a *key map* is used to define the mapping from a primary key in the 
    *from_metaclass* to a foreign key in the *to_metaclass*. The *phrase* is
    used to specify the direction when the association is reflexive.
    
    In addition, links also specify cardinality constraints via the *many*
    and *conditional* attributes.
    '''
    from_metaclass = None
    rel_id = None
    to_metaclass = None
    phrase = None
    key_map = None
    conditional = None
    many = None
    
    def __init__(self, from_metaclass, rel_id, to_metaclass, phrase='',
                 conditional=False, many=False):
        if isinstance(rel_id, int):
            rel_id = 'R%d' % rel_id
        
        self.from_metaclass = from_metaclass
        self.rel_id = rel_id
        self.to_metaclass = to_metaclass
        self.key_map = dict()
        self.phrase = phrase
        self.conditional = conditional
        self.many = many

    @property
    def cardinality(self):
        '''
        Obtain the cardinality string.
        
        Example: '1C' for a conditional link with a single instance [0..1]
                 'MC' for a link with any number of instances [0..*]
                 'M'  for a more than one instance [1..*]
                 'M'  for a link with exactly one instance [1]
        '''
        if self.many:
            s = 'M'
        else:
            s = '1'
            
        if self.conditional:
            s += 'C'
            
        return s
    
    @property
    def kind(self):
        '''
        Obtain the resulting kind when the link is navigated.
        '''
        return self.to_metaclass.kind

    def connect(self, instance, another_instance, check=True):
        '''
        Connect an *instance* to *another_instance*.
        
        Optionally, disable any cardinality *check* that would prevent the two
        instances from being connected.
        '''
        if instance not in self:
            self[instance] = xtuml.OrderedSet()
        
        if another_instance in self[instance]:
            return True
        
        if self[instance] and not self.many and check:
            return False  

        self[instance].add(another_instance)
        return True
        
    def disconnect(self, instance, another_instance):
        '''
        Disconnect an *instance* from *another_instance*.
        '''
        if instance not in self:
            return False
        
        if another_instance not in self[instance]: 
            return False

        self[instance].remove(another_instance)
        return True
        
    def navigate(self, instance):
        '''
        Navigate from *instance* across the link.
        '''
        if instance in self:
            return self[instance]
        else:
            return set()
        
    def navigate_one(self, instance):
        '''
        Navigate from *instance* across the link.
        '''
        return next(iter(self.navigate(instance)), None)
        
    def compute_lookup_key(self, from_instance):
        '''
        Compute the lookup key for an instance, i.e. a foreign key that
        can be used to identify an instance at the end of the link.
        '''
        kwargs = dict()
        for attr, other_attr in self.key_map.items():
            if _is_null(from_instance, attr):
                return None
            
            if attr in from_instance.__dict__:
                kwargs[other_attr] = from_instance.__dict__[attr]
            else:
                kwargs[other_attr] = getattr(from_instance, attr)

        return frozenset(tuple(kwargs.items()))

    def compute_index_key(self, to_instance):
        '''
        Compute the index key that can be used to identify an instance
        on the link.
        '''
        kwargs = dict()
        for attr in self.key_map.values():
            if _is_null(to_instance, attr):
                return None
            
            if attr in to_instance.__dict__:
                kwargs[attr] = to_instance.__dict__[attr]
            else:
                kwargs[attr] = getattr(to_instance, attr)

        return frozenset(tuple(kwargs.items()))

    
class QuerySet(xtuml.OrderedSet):
    '''
    An ordered set which holds instances that match queries.
    '''
    @property
    def first(self):
        '''
        Obtain the first element in the set.
        '''
        if len(self):
            return next(iter(self))
    
    @property
    def last(self):
        '''
        Obtain the last element in the set.
        '''
        if len(self):
            return next(reversed(self))


class Class(object):
    '''
    A class that all instances created by a metaclass inherits from. 
    
    **Note**: Accesses to attributes, e.g. getattr/setattr, on these objects
    are case insensitive.
    '''
    def __add__(self, other):
        assert isinstance(other, Class)
        return QuerySet([self, other])

    def __sub__(self, other):
        assert isinstance(other, Class)
        if self == other: return QuerySet()
        else: return QuerySet([self])

    def __getattr__(self, name):
        uname = name.upper()
        for attr, _ in get_metaclass(self).attributes:
            if attr.upper() != uname :
                continue
            
            if attr in self.__dict__:
                return self.__dict__[attr]
            else:
                return object.__getattribute__(self, attr)
        
        return object.__getattribute__(self, name)
    
    def __setattr__(self, name, value):
        uname = name.upper()
        for attr, _ in get_metaclass(self).attributes:
            if attr.upper() != uname :
                continue

            if attr in self.__dict__:
                self.__dict__[attr] = value
            else:
                return object.__setattr__(self, attr, value)
        
        self.__dict__[name] = value
        
    def __delattr__(self, name):
        uname = name.upper()
        for name in self.__dict__:
            if uname == name.upper():
                break

        del self.__dict__[name]
    
    def __str__(self):
        values = list()
        for attr, ty in get_metaclass(self).attributes:
            value = getattr(self, attr)
            value = xtuml.serialize_value(value, ty)
            values.append('%s=%s' % (attr, value))
        
        return '%s(%s)' % (self.__class__.__name__, ', '.join(values))


# Backwards compatibility with older versions of pyxtuml
BaseObject = Class


class MetaClass(object):
    '''
    A metaclass contain metadata for instances, e.g. what attributes are 
    available, what thier types are, and so on.
    
    In addition, each metaclass also handle allocations of instances.
    '''
    metamodel = None
    kind = None
    attributes = None
    referential_attributes = None
    identifying_attributes = None
    links = None
    indices = None
    clazz = None
    storage = None
    
    def __init__(self, kind, metamodel=None):
        self.metamodel = metamodel
        self.kind = kind
        self.attributes = list()
        self.referential_attributes = set()
        self.identifying_attributes = set()
        self.indices = dict()
        self.links = dict()
        self.storage = list()
        self.clazz = type(str(kind), (Class,), dict(__metaclass__=self))
        
    def __call__(self, *args, **kwargs):
        '''
        Create and return a new instance using the metaclass constructor.
        '''
        return self.new(*args, **kwargs)
        
    @property
    def attribute_names(self):
        '''
        Obtain an ordered list of all attribute names.
        '''
        return [name for name, _ in self.attributes]

    def attribute_type(self, attribute_name):
        '''
        Obtain the type of an attribute.
        '''
        attribute_name = attribute_name.upper()
        for name, ty in self.attributes:
            if name.upper() == attribute_name:
                return ty
    
    def add_link(self, metaclass, rel_id, phrase, conditional, many):
        '''
        Add a new link from *self* to *metaclass*.
        '''
        link = Link(self, rel_id, metaclass, phrase, conditional, many)
        key = (metaclass.kind.upper(), rel_id, phrase)
        self.links[key] = link

        return link
            
    def append_attribute(self, name, type_name):
        '''
        Append an attribute with a given *name* and *type name* at the end of
        the list of attributes.
        '''
        attr = (name, type_name)
        self.attributes.append(attr)
        
    def insert_attribute(self, index, name, type_name):
        '''
        Insert an attribute with a given *name* and *type name* at some *index*
        in the list of attributes.
        '''
        attr = (name, type_name)
        self.attributes.insert(index, attr)
        
    def delete_attribute(self, name):
        '''
        Delete an attribute with a given *name* from the list of attributes.
        '''
        for idx, attr in enumerate(self.attributes):
            attr_name, _ = attr
            if attr_name == name:
                del self.attributes[idx]
                return
        
    def default_value(self, type_name):
        '''
        Obtain the default value for some *type name*.
        '''
        uname = type_name.upper()
        if   uname == 'BOOLEAN':
            return False
            
        elif uname == 'INTEGER':
            return 0
            
        elif uname == 'REAL':
            return 0.0
            
        elif uname == 'STRING':
            return ''
            
        elif uname == 'UNIQUE_ID':
            if self.metamodel:
                return next(self.metamodel.id_generator)
            else:
                return None
        else:
            raise MetaException("Unknown type named '%s'" % type_name)
        
    def new(self, *args, **kwargs):
        '''
        Create and return a new instance.
        '''
        inst = self.clazz()
        self.storage.append(inst)
        
        # set all attributes with an initial default value
        referential_attributes = dict()
        for name, ty in self.attributes:
            if name not in self.referential_attributes:
                value = self.default_value(ty)
                setattr(inst, name, value)
            
        # set all positional arguments
        for attr, value in zip(self.attributes, args):
            name, ty = attr
            if name not in self.referential_attributes:
                setattr(inst, name, value)
            else:
                referential_attributes[name] = value
            
        # set all named arguments
        for name, value in kwargs.items():
            if name not in self.referential_attributes:
                setattr(inst, name, value)
            else:
                referential_attributes[name] = value
        
        if not referential_attributes:
            return inst
        
        # batch relate referential attributes 
        for link in self.links.values():
            if set(link.key_map.values()) - set(referential_attributes.keys()):
                continue
             
            kwargs = dict()
            for key, value in link.key_map.items():
                kwargs[key] = referential_attributes[value]
            
            if not kwargs:
                continue
            
            for other_inst in link.to_metaclass.query(kwargs):
                relate(other_inst, inst, link.rel_id, link.phrase)
        
        for name, value in referential_attributes.items():
            if getattr(inst, name) != value:
                logger.warning('unable to assign %s to %s', name, inst)
                
        return inst

    def clone(self, instance):
        '''
        Create a shallow clone of an *instance*.
        
        **Note:** the clone and the original instance **does not** have to be
        part of the same metaclass. 
        '''
        args = list()
        for name, _ in get_metaclass(instance).attributes:
            value = getattr(instance, name)
            args.append(value)
            
        return self.new(*args)
    
    def delete(self, instance, disconnect=True):
        '''
        Delete an *instance* from the instance pool and optionally *disconnect*
        it from any links it might be connected to. If the *instance* is not
        part of the metaclass, a *MetaException* is thrown.
        '''
        if instance in self.storage:
            self.storage.remove(instance)
        else:
            raise DeleteException("Instance not found in the instance pool")

        if not disconnect:
            return
        
        for link in self.links.values():
            if instance not in link:
                continue
        
            for other in link[instance]:
                unrelate(instance, other, link.rel_id, link.phrase)
        
    def select_one(self, *args):
        '''
        Select a single instance from the instance pool. Query operators such as
        where_eq(), order_by() or filter functions may be passed as optional
        arguments.
        '''
        s = apply_query_operators(self.storage, args)
        return next(iter(s), None)

    def select_many(self, *args):
        '''
        Select several instances from the instance pool. Query operators such as
        where_eq(), order_by() or filter functions may be passed as optional
        arguments.
        '''
        s = apply_query_operators(self.storage, args)
        if isinstance(s, QuerySet):
            return s
        else:
            return QuerySet(s)

    def _find_assoc_links(self, kind, rel_id, phrase=''):
        key = (kind.upper(), rel_id, phrase)
        for link in self.links.values():
            if link.rel_id != rel_id or link.phrase != phrase:
                continue
            
            metaclass = self.metamodel.find_metaclass(link.kind)
            if key in metaclass.links:
                return link, metaclass.links[key]
        
        raise UnknownLinkException(self.kind, kind, rel_id, phrase)
    
    def navigate(self, inst, kind, rel_id, phrase=''):
        '''
        Navigate across a link with some *rel_id* and *phrase* that yields
        instances of some *kind*.
        '''
        key = (kind.upper(), rel_id, phrase)
        if key in self.links:
            link = self.links[key]
            return link.navigate(inst)
        
        link1, link2 = self._find_assoc_links(kind, rel_id, phrase)
        inst_set = xtuml.OrderedSet()
        for inst in link1.navigate(inst):
            inst_set |= link2.navigate(inst)
        
        return inst_set
    
    def query(self, dictonary_of_values):
        '''
        Query the instance pool for instances with attributes that match a given
        *dictonary of values*.
        '''
        return WhereEqual(dictonary_of_values)(self.storage)
    

class NavChain(object):
    '''
    A navigation chain initializes a navigation from one or more instances.
    Navigation may be syntactically cascaded in several ways:
    
       res = NavChain(inst).nav('X', 'R100', 'phrase').nav('Y', 101)

    or using an OAL/RSL inspired syntax:
    
       res = NavChain(inst).X[100, 'phrase'].Y[101](lamda x: <filter expression>)
    '''
    
    def __init__(self, handle):
        if handle is None:
            handle = []
            
        elif isinstance(handle, Class):
            handle = [handle]
        
        elif not isinstance(handle, collectionsAbc.Iterable):
            raise MetaException("Unable to navigate across '%s'" % type(handle))
        
        self.handle = handle
        self._kind = None
        
    def nav(self, kind, relid, phrase=''):
        self.handle = NavChain._nav(self.handle, kind, relid, phrase)
        return self
    
    @staticmethod
    def _nav(handle, kind, rel_id, phrase):
        if isinstance(rel_id, int):
            rel_id = 'R%d' % rel_id
    
        for inst in iter(handle):
            metaclass = get_metaclass(inst)
            for result in metaclass.navigate(inst, kind, rel_id, phrase):
                yield result
            
    def __getattr__(self, kind):
        '''
        The navigation chain specified a *kind*, e.g.
        >>> chain.X
        '''
        self._kind = kind
        return self
    
    def __getitem__(self, args):
        '''
        The navigation chain specified a link, e.g. the rel_id and phrase, e.g.
        >>> chain[100, 'phrase']
        '''
        if not isinstance(args, tuple):
            args = (args, '')
        
        relid, phrase = args
        
        return self.nav(self._kind, relid, phrase)

    def __call__(self, *args):
        '''
        The navigation chain is invoked. Query operators such as where_eq(), 
        order_by() or filter functions may be passed as optional arguments, e.g.

        >>> chain(lambda selected: selected.Name == 'test')
        '''
        handle = self.handle or list()
        handle = apply_query_operators(handle, args)
        if isinstance(handle, QuerySet):
            return handle
        else:
            return QuerySet(handle)
    
    
class NavOneChain(NavChain):
    '''
    A navigation chain that yeilds an instance, or None.
    '''
    def __call__(self, *args):
        handle = self.handle or list()
        handle = apply_query_operators(handle, args)
        return next(iter(handle), None)


def navigate_one(instance):
    '''
    Initialize a navigation from one *instance* to another across a one-to-one
    association.
    
    The resulting query will return an instance or None.
    
    Usage example:
    
    >>> from xtuml import navigate_one as one
    >>> m = xtuml.load_metamodel('db.sql')
    >>> inst = m.select_any('My_Modeled_Class')
    >>> other_inst = one(inst).Some_Other_Class[4]()
    
    The syntax is somewhat similar to the action language used in BridgePoint.
    The same semantics would be expressed in BridgePoint as::
    
        select any inst from instances of My_Modeled_Class;
        select one other_inst related by inst->Some_Other_Class[R4];
    
    **Note:** If the navigated association is reflexive, a phrase must be 
    provided, e.g.
    
    >>> other_inst = one(inst).Some_Other_Class[4, 'some phrase']()
    '''
    return navigate_any(instance)


def navigate_any(instance_or_set):
    '''
    Initialize a navigation from an instance, or a set of instances, to 
    associated instances across a one-to-many or many-to-many association.

    The resulting query will return an instance or None.
    '''
    return NavOneChain(instance_or_set)


def navigate_many(instance_or_set):
    '''
    Initialize a navigation from an instance, or a set of instances, to 
    associated instances across a one-to-many or many-to-many association.
    
    The resulting query will return a set of instances.
    '''
    return NavChain(instance_or_set)


def navigate_subtype(supertype, rel_id):
    '''
    Perform a navigation from *supertype* to its subtype across *rel_id*. The
    navigated association must be modeled as a subtype-supertype association.
    
    The return value will an instance or None.
    '''
    if not supertype:
        return
    
    if isinstance(rel_id, int):
        rel_id = 'R%d' % rel_id

    metaclass = get_metaclass(supertype)
    for kind, rel_id_candidate, _ in metaclass.links:
        if rel_id != rel_id_candidate:
            continue
        
        subtype = navigate_one(supertype).nav(kind, rel_id)()
        if subtype:
            return subtype


class WhereEqual(dict):
    '''
    Helper class to create a dictonary of values for queries using
    python keyword arguments to *where_eq()*
    '''
    def __call__(self, s):
        items = collections.deque(self.items())
        for inst in iter(s):
            for name, value in iter(items):
                if getattr(inst, name) != value:
                    break
            else:
                yield inst


def where_eq(**kwargs):
    '''
    Return a where-clause that filters out instances based on named 
    keywords.
    
    Usage example:
    
    >>> from xtuml import where_eq as where
    >>> m = xtuml.load_metamodel('db.sql')
    >>> inst = m.select_any('My_Modeled_Class', where(My_Number=5))
    '''
    return WhereEqual(kwargs)


def sort_reflexive(set_of_instances, rel_id, phrase):
    '''
    Sort a *set of instances* in the order they appear across a conditional and
    reflexive association. The first instance in the resulting ordered set is
    **not** associated to an instance across the given *phrase*.
    '''
    if not isinstance(set_of_instances, QuerySet):
        raise MetaException('The collection to sort must be a QuerySet')
    
    if not set_of_instances.first:
        return QuerySet()
    
    if isinstance(rel_id, int):
        rel_id = 'R%d' % rel_id
    
    # Figure out the phrase in the other direction
    metaclass = get_metaclass(set_of_instances.first)
    for link in metaclass.links.values():
        if link.to_metaclass != metaclass:
            continue
        
        if link.rel_id != rel_id:
            continue

        if link.phrase == phrase:
            continue

        other_phrase = link.phrase
        break
    else:
        raise UnknownLinkException(metaclass.kind, metaclass.kind, rel_id, phrase)
    
    first_filt = lambda sel: not navigate_one(sel).nav(metaclass.kind, rel_id, phrase)()
    first_instances = list(filter(first_filt, set_of_instances))
    if not first_instances:
        #the instance sequence is recursive, start anywhere
        first_instances = [set_of_instances.first]
    
    def sequence_generator():
        for first in first_instances:
            inst = first
            while inst:
                if inst in set_of_instances:
                    yield inst
                inst = navigate_one(inst).nav(metaclass.kind, rel_id, other_phrase)()
                if inst is first:
                    break
                
    return QuerySet(sequence_generator())

    
def _find_link(inst1, inst2, rel_id, phrase):
    '''
    Find links that correspond to the given arguments.
    '''
    metaclass1 = get_metaclass(inst1)
    metaclass2 = get_metaclass(inst2)

    if isinstance(rel_id, int):
        rel_id = 'R%d' % rel_id
        
    for ass in metaclass1.metamodel.associations:
        if ass.rel_id != rel_id:
            continue

        if (ass.source_link.from_metaclass.kind == metaclass1.kind and
            ass.source_link.to_metaclass.kind == metaclass2.kind and
            ass.source_link.phrase == phrase):
            return inst1, inst2, ass

        if (ass.target_link.from_metaclass.kind == metaclass1.kind and
            ass.target_link.to_metaclass.kind == metaclass2.kind and
            ass.target_link.phrase == phrase):
            return inst2, inst1, ass

    raise UnknownLinkException(metaclass1.kind, metaclass2.kind, rel_id, phrase)


def relate(from_instance, to_instance, rel_id, phrase=''):
    '''
    Relate *from_instance* to *to_instance* across *rel_id*. For reflexive
    association, a *phrase* indicating the direction must also be provided.
    
    The two instances are related to each other by copying the identifying 
    attributes from the instance on the TO side of a association to the instance
    n the FROM side. Updated values which affect existing associations are 
    propagated. A set of all affected instances will be returned.
    '''
    if None in [from_instance, to_instance]:
        return False

    inst1, inst2, ass = _find_link(from_instance, to_instance, rel_id, phrase)
    if not ass.source_link.connect(inst1, inst2):
        raise RelateException(from_instance, to_instance, rel_id, phrase)

    if not ass.target_link.connect(inst2, inst1):
        raise RelateException(from_instance, to_instance, rel_id, phrase)
    
    return True


def unrelate(from_instance, to_instance, rel_id, phrase=''):
    '''
    Unrelate *from_instance* from *to_instance* across *rel_id*. For reflexive
    associations, a *phrase* indicating the direction must also be provided.
    
    The two instances are unrelated from each other by reseting the identifying
    attributes on the FROM side of the association. Updated values which affect
    existing associations are propagated. A set of all affected instances will
    be returned.
    '''
    if None in [from_instance, to_instance]:
        return False
    
    inst1, inst2, ass = _find_link(from_instance, to_instance, rel_id, phrase)
    if not ass.source_link.disconnect(inst1, inst2):
        raise UnrelateException(from_instance, to_instance, rel_id, phrase)

    if not ass.target_link.disconnect(inst2, inst1):
        raise UnrelateException(from_instance, to_instance, rel_id, phrase)
        
    return True


def get_metaclass(class_or_instance):
    '''
    Get the metaclass for a *class_or_instance*.
    '''
    if isinstance(class_or_instance, Class):
        return class_or_instance.__metaclass__
    
    elif issubclass(class_or_instance, Class):
        return class_or_instance.__metaclass__
    
    raise MetaException("the provided argument is not an xtuml class or instance")
    

def get_metamodel(class_or_instance):
    '''
    Get the metamodel in which a *class_or_instance* was created.
    '''
    return get_metaclass(class_or_instance).metamodel


def delete(instance, disconnect=True):
    '''
    Delete an *instance* from its metaclass instance pool and optionally
    *disconnect* it from any links it might be connected to.
    '''
    if not isinstance(instance, Class):
        raise DeleteException("the provided argument is not an xtuml instance")
            
    return get_metaclass(instance).delete(instance, disconnect)


def cardinality(instance_or_set):
    '''
    Get the cardinality of an *instance_or_set*.
    '''
    if not instance_or_set:
        return 0
    
    if isinstance(instance_or_set, Class):
        return 1
    
    return len(instance_or_set)


class OrderBy(list):
    '''
    Helper class to create a tuple of key values for sorting an
    instance set.
    '''
    reverse = False

    def __init__(self, attrs, reverse=False):
        list.__init__(self, attrs)
        self.reverse = reverse

    def __call__(self, s):
        key = lambda el: [getattr(el, name) for name in self]
        return sorted(s, key=key, reverse=self.reverse)


def order_by(*attrs):
    '''
    Return a query ordering operator that will order an instance
    set based on attribute names passed. When ordering on multiple
    attributes is specified, the set will be sorted by the first
    attribute and then within each value of this, by the second
    attribute and so on.
    
    Usage example:
    
    >>> from xtuml import order_by
    >>> m = xtuml.load_metamodel('db.sql')
    >>> inst = m.select_many('My_Class', order_by('Name', 'Number'))
    '''
    return OrderBy(attrs, reverse=False)


def reverse_order_by(*attrs):
    '''
    Return a reversed query ordering operator with the same behavior as
    order_by() but reversed order.
    
    Usage example:
    
    >>> from xtuml import reverse_order_by
    >>> m = xtuml.load_metamodel('db.sql')
    >>> inst = m.select_many('My_Class', reverse_order_by('Name', 'Number'))
    '''
    return OrderBy(attrs, reverse=True)
    

class MetaModel(object):
    '''
    A metamodel contains metaclasses with associations between them.
    
    **Note:** All identifiers, e.g. attributes, association ids, key letters 
    (the kind or name of a class), are case **insensitive**.
    '''
    metaclasses = None
    associations = None
    id_generator = None
    
    def __init__(self, id_generator=None):
        '''
        Create a new, empty metamodel. Optionally, specify an id generator
        used to obtain unique identifiers.
        '''
        if id_generator is None:
            id_generator = xtuml.UUIDGenerator()
        
        self.metaclasses = dict()
        self.associations = list()
        self.id_generator = id_generator
    
    @property
    def instances(self):
        '''
        Obtain a sequence of all instances in the metamodel.
        '''
        for metaclass in self.metaclasses.values():
            for inst in metaclass.storage:
                yield inst
    
    def define_class(self, kind, attributes, doc=''):
        '''
        Define a new class in the metamodel, and return its metaclass.
        '''
        ukind = kind.upper()
        if ukind in self.metaclasses:
            raise MetaModelException('A class with the name %s is already defined' % kind)

        metaclass = MetaClass(kind, self)
        for name, ty in attributes:
            metaclass.append_attribute(name, ty)
            
        self.metaclasses[ukind] = metaclass
        
        return metaclass

    def find_class(self, kind):
        '''
        Find a class of some *kind* in the metamodel.
        '''
        return self.find_metaclass(kind).clazz

    def find_metaclass(self, kind):
        '''
        Find a metaclass of some *kind* in the metamodel.
        '''
        ukind = kind.upper()
        if ukind in self.metaclasses:
            return self.metaclasses[ukind]
        else:
            raise UnknownClassException(kind)

    def new(self, kind, *args, **kwargs):
        '''
        Create and return a new instance in the metamodel of some *kind*.
        
        Optionally, initial attribute values may be assigned to the new instance
        by passing them as positional or keyword arguments. Positional arguments
        are assigned in the order in which they appear in the metaclass.
        '''
        metaclass = self.find_metaclass(kind)
        return metaclass.new(*args, **kwargs)
        
    def clone(self, instance):
        '''
        Create a shallow clone of an *instance*.
        
        **Note:** the clone and the original instance **does not** have to be
        part of the same metaclass. 
        '''
        metaclass = get_metaclass(instance)
        metaclass = self.find_metaclass(metaclass.kind)
        return metaclass.clone(instance)
            
    def define_association(self, rel_id, source_kind, source_keys, source_many,
                           source_conditional, source_phrase, target_kind, 
                           target_keys, target_many, target_conditional, 
                           target_phrase):
        '''
        Define and return an association from one kind of class (the source 
        kind) to some other kind of class (the target kind).
        '''
        if isinstance(rel_id, int):
            rel_id = 'R%d' % rel_id
            
        source_metaclass = self.find_metaclass(source_kind)
        target_metaclass = self.find_metaclass(target_kind)

        source_link = target_metaclass.add_link(source_metaclass, rel_id,
                                                many=source_many,
                                                phrase=target_phrase,
                                                conditional=source_conditional)
                
        target_link = source_metaclass.add_link(target_metaclass, rel_id,
                                                many=target_many,
                                                phrase=source_phrase,
                                                conditional=target_conditional)
        
        ass = Association(rel_id,
                          source_keys, source_link,
                          target_keys, target_link)
        
        source_link.key_map = dict(zip(source_keys, target_keys))
        target_link.key_map = dict(zip(target_keys, source_keys))
        
        self.associations.append(ass)

        return ass
        
    def define_unique_identifier(self, kind, name, *named_attributes):
        '''
        Define a unique identifier for some *kind* of class based on its
        *named attributes*.
        '''
        if not named_attributes:
            return
        
        if isinstance(name, int):
            name = 'I%d' % name
        
        metaclass = self.find_metaclass(kind)
        metaclass.indices[name] = tuple(named_attributes)
        metaclass.identifying_attributes |= set(named_attributes)

    def select_many(self, kind, *args):
        '''
        Query the metamodel for a set of instances of some *kind*. Query
        operators such as where_eq(), order_by() or filter functions may be
        passed as optional arguments.
        
        Usage example:
        
        >>> m = xtuml.load_metamodel('db.sql')
        >>> inst_set = m.select_many('My_Class', lambda sel: sel.number > 5)
        '''
        metaclass = self.find_metaclass(kind)
        return metaclass.select_many(*args)
    
    def select_one(self, kind, *args):
        '''
        Query the metamodel for a single instance of some *kind*. Query
        operators such as where_eq(), order_by() or filter functions may be
        passed as optional arguments.
        
        Usage example:
        
        >>> m = xtuml.load_metamodel('db.sql')
        >>> inst = m.select_one('My_Class', lambda sel: sel.name == 'Test')
        '''
        metaclass = self.find_metaclass(kind)
        return metaclass.select_one(*args)
    
    # Backwards compatibility with older versions of pyxtuml
    select_any = select_one

    def is_consistent(self):
        '''
        Check the metamodel for integrity violations.
        '''
        if xtuml.check_association_integrity(self):
            return False
        
        return xtuml.check_uniqueness_constraint(self) == 0

                
